## yapCAD polyline and polygon sampling and drawing example

print("example5.py -- yapCAD computational geometry and DXF drawing example")
print("create, sample, and draw simple polylines and polygons")

print("""

In this demo we create simple polyline/polygon figures by regularly
sampling circles, then addign those points to a list.  By repeating
the first point as the last point, we close the figure.

We then use drawable objects to draw the figures, and sample each
polyline to create a series of circles drawn on top.  The circles
increase in size to show the direction of increasing sample parameter.
""")

from yapcad.ezdxf_drawable import *
from yapcad.geom import *

#set up DXF rendering
d=ezdxfDraw()
filename="example5-out"
print("Output file name is {}.dxf".format(filename))
d.filename = filename

d.layer = 'DOCUMENTATION'

d.draw_text("yapCAD", point(5,15),\
            attr={'style': 'OpenSans-Bold',
                  'height': 1.5})

d.draw_text("example5.py",
            point(5,12))
d.draw_text("constructing simple polys",
            point(5,10))
d.draw_text("and sampling outlines",
            point(5,8.5))
d.layer = False
# make circles centered at -10.0,10.0
a = point(-10.0,10.0)

# make circles centered at 10.0,-10.0
b = point(10.0,-10.0)

## sample points from the circles to make regular polygons

## this will be a list of polys, which is to say a list of lists of
## points
polys = []

## NOTE: a simple polyline is just a list of points
for i in range(4,7):
    poly = []
    circ1 = arc(a,i**1.4)
    for j in range(i):
        x=j/i
        poly.append(samplearc(circ1,x)) # sample a point and add it
    poly.append(poly[0]) # make the figure closed
    polys.append(poly) # add the finished poly to the list

for i in range(7,10):
    poly = []
    circ2 = arc(b,(i*0.7)**1.4)
    for j in range(i):
        x=j/i
        poly.append(samplearc(circ2,x)) # sample a point and add it
    poly.append(poly[0]) # make the figure closed
    polys.append(poly) # add the finished poly to the list


## draw the geometry list of polys
d.draw(polys)

# Sample spome points along each poly and use the centers to draw
# circles of increasing size.  This allows the visual confirmation of
# even spacing and shows the counter-clockwise progression of samples

d.layer = 'DRILLS' #select drills layer for circles

i = 4
for ply in polys:
    for j in range(i*3):
        p = samplepoly(ply,j/(i*3))
        d.draw(arc(p,0.2+j/(i*10)))
    i = i+1
    
d.display()



## yapCAD boolean operations geometry example

from yapcad.geom import *
from yapcad.poly import *
from yapcad.combine import *
import examples.example8 as example8

docGeomList=[]

def makeJigsawPiece(width=100,height=50,diam=30,frac=0.80,upper=False,poly=False):
    rect = makeRoundRect(width,height,5.0)
    rect = rect.translate(point(0,-height/2),poly=True)
    circ1 = makeCircle(point(-width/4,(diam/2)*frac),diam/2)
    #circ2 = makeCircle(point(width/4,-(diam/2)*frac),diam/2)
    circ2 = makeRoundRect(diam,diam,20.0,center=point(width/4,-(diam/2)*frac))

    docGeomList.append(circ1.geom())
    docGeomList.append(circ2.geom())
    bln = Boolean('union',[circ1,rect])
    bln = Boolean('difference',[bln,circ2])
    docGeomList.append(bln.geom())
    if poly:
        return bln
    return bln.geom()

def drawLegend(d):
    ## Put some documentary text on the drawing
    d.layer = 'DOCUMENTATION'

    att = {'style': 'OpenSans-Bold',
           'height': 2.5}
    
    d.draw_text("yapCAD", point(5,15), attr = att)

    d.draw_text("example12.py", point(5,11), attr=att)
    d.draw_text("complex boolean operations on Polygon() instances",
                point(5,7),attr=att)
    d.layer = False # back to default layer


def geometry():

    logopoly = example8.flower(returnPoly=True)
    jigsaw=makeJigsawPiece(poly=True)
    bigc= makeCircle(point(0,0,0),20)
    nb = Boolean('union',[logopoly,bigc])
    nb = nb.translate(point(0,-50),poly=True)
    docGeomList.append(nb.geom())

    nb = Boolean('difference',[jigsaw,nb])
    
    gl = nb.geom()
    gl2 = []

    for i in range(100):
        u = i/99.0
        rad = 1.0 + u*2
        
        p = nb.sample(u)
        c = arc(p,rad)
        gl2.append(c)
        

    return gl,gl2


def testAndDraw(dd):
    geom,geom2=geometry()

    if not renderOgl:
        dd.layer = 'DRILLS'
        dd.linecolor = False

    else:
        dd.linecolor = 'aqua'
    dd.draw(geom2)

    if not renderOgl:
        dd.layer = 'DOCUMENTATION'
        dd.linecolor = False
    else:
        dd.linecolor = 'yellow'
    dd.draw(docGeomList)

    if not renderOgl:
        dd.layer = 'PATHS'
        
    dd.linecolor = 'white'
    dd.draw(geom)

    dd.display()
    

if __name__ == "__main__":
    import sys
    renderOgl = False
    filename="example12-out"
    print("example12.py -- yapCAD boolean operations demonstration")

    oglarg= ("pyglet","opengl","OpenGL")
    dxfarg= ("ezdxf","dxf","DXF")
    if len(sys.argv) > 1:
        if sys.argv[1] in oglarg:
            renderOgl=True
        elif sys.argv[1] in dxfarg:
            renderOgl=False
        else:
            print("This is an example of a complex boolean operation on Polygon() and")
            print("Boolean() instances. ")
            print("syntax: $ python3 {} <rendertype> [filename]".format(sys.argv[0]))
            print("    where <rendertype> is one of {} for OpenGL".format(oglarg))
            print("    or one of {} for DXF".format(dxfarg))
            print("    For DXF, you can optionally specify [filename].dxf as the output file")
            quit()
    if len(sys.argv) > 2 and renderOgl==False:
        filename = sys.argv[2]+".dxf"
    dd = []
    if renderOgl:
        print("OpenGL rendering selected")
        from yapcad.pyglet_drawable import *
        dd=pygletDraw()
        dd.cameradist=170.0
        
    else:
        print("DXF rendering selected")
        from yapcad.ezdxf_drawable import *
        #set up DXF rendering
        dd=ezdxfDraw()
        dd.filename = filename
        drawLegend(dd)
    print("rendering...")
    
    testAndDraw(dd)
    print("done")



## yapCAD 3D geometry example

from yapcad.geom import *
import math

## utility function, convert lat, lon, rad to cartesian coordinates
def lat_lon_rad(lat,lon,rad):
    if lat == 90:
        return [0,0,rad,1]
    elif lat == -90:
        return [0,0,-rad,1]
    else:
        latr = (((lat+90)%180)-90)*pi2/360.0
        lonr = (lon%360)*pi2/360.0
    
        smallrad = math.cos(latr)*rad
        z = math.sin(latr)*rad
        x = math.cos(lonr)*smallrad
        y = math.sin(lonr)*smallrad
        return [x,y,z,1]

## icosohedron-specific function, generate intial geometry
def makeIcoPoints(center,radius):
    points = []
    normals = []
    p = lat_lon_rad(90,0,radius)
    n = scale3(p,1.0/mag(p))
    points.append(p)
    normals.append(n)
    for i in range(10):
        sgn = 1
        if i%2 == 0:
            sgn =-1
        lat = math.atan(0.5)*360.0*sgn/pi2
        lon = i*36.0
        p = lat_lon_rad(lat,lon,radius)
        n = scale3(p,1.0/mag(p))
        points.append(p)
        normals.append(n)

    p = lat_lon_rad(-90,0,radius)
    n = scale3(p,1.0/mag(p))
    points.append(p)
    normals.append(n)
    return list(map( lambda x: add(x,center),points)),normals

# face indices for icosahedron
icaIndices = [ [1,11,3],[3,11,5],[5,11,7],[7,11,9],[9,11,1],
               [2,1,3],[2,3,4],[4,3,5],[4,5,6],[6,5,7],[6,7,8],[8,7,9],[8,9,10],[10,9,1],[10,1,2],
               [0,2,4],[0,4,6],[0,6,8],[0,8,10],[0,10,2] ]


def addVertex(nv,nn,verts,normals):
    for i in range(len(verts)):
        if vclose(nv,verts[i]):
            return i,verts,normals
    verts.append(nv)
    normals.append(nn)
    return len(verts)-1,verts,normals

def subdivide(f,verts,normals,rad):
    ind1 = f[0]
    ind2 = f[1]
    ind3 = f[2]
    v1 = verts[ind1]
    v2 = verts[ind2]
    v3 = verts[ind3]
    n1 = normals[ind1]
    n2 = normals[ind2]
    n3 = normals[ind3]
    va = add(v1,v2)
    vb = add(v2,v3)
    vc = add(v3,v1)
    ma = rad/mag(va)
    mb = rad/mag(vb)
    mc = rad/mag(vc)
    va = scale3(va,ma)
    vb = scale3(vb,mb)
    vc = scale3(vc,mc)
    
    na = add(n1,n2)
    na = scale3(na,1.0/mag(na))
    nb = add(n2,n3)
    nb = scale3(nb,1.0/mag(nb))
    nc = add(n3,n1)
    nc = scale3(nc,1.0/mag(nc))

    inda,verts,normals = addVertex(va,na,verts,normals)
    indb,verts,normals = addVertex(vb,nb,verts,normals)
    indc,verts,normals = addVertex(vc,nc,verts,normals)

    f1 = [ind1,inda,indc]
    f2 = [inda,ind2,indb]
    f3 = [indb,ind3,indc]
    f4 = [inda,indb,indc]

    return verts,normals, [f1,f2,f3,f4]

    

# make the sphere, return a surface representation
def sphere(diameter,center=point(0,0,0),depth=2):
    rad = diameter/2
    verts,normals = makeIcoPoints(center,rad)
    faces = icaIndices
    
    for i in range(depth):
        ff = []
        for f in faces:
            verts, norms, newfaces = subdivide(f,verts,normals,rad)
            ff+=newfaces
        faces = ff
                      
    
    return [verts,normals,faces]
    

        
if __name__ == "__main__":
    from yapcad.pyglet_drawable import *
    print("example9.py -- yapCAD 3D geometry demonstration")
    print("""
This is a demonstration of the creation and rendering of
three-dimensional geometry using yapCAD and the pyglet openGL draing
engine.

In this example we create an icosohedron centered at the orign and
tesellate it spherically""")

    dd=pygletDraw()
    dd.linecolor = 'white'

    verts,normals,faces = sphere(50.0,point(0,0,0),3)
    
    for f in faces:
        dd.draw(line(verts[f[0]],
                     verts[f[1]]))
        dd.draw(line(verts[f[1]],
                     verts[f[2]]))
        dd.draw(line(verts[f[2]],
                     verts[f[0]]))

    dd.draw_surface(verts,normals,faces)
    dd.linecolor = 'aqua'
    dd.polystyle = 'points'
    dd.pointstyle = 'xo'
    dd.draw(verts)
    dd.display()



## simple yapCAD framework for openGL drawing using pyglet
## package
## Copyright (c) 2020 Richard W. DeVaul
## Copyright (c) 2020 yapCAD contributors
## All rights reserved
## See licensing terms here: https://github.com/rdevaul/yapCAD/blob/master/LICENSE

import math
import pyglet
import pyglet.font as font
import pyglet.gl as gl
import pyglet.graphics as graphics

from yapcad.geom import *
import yapcad.drawable as drawable

## openGL utility functions
def vec(*args):
    return (gl.GLfloat * len(args))(*args)

## HTML document, instructions for user interaction
yapCAD_legend="""
<font face="OpenSans, Geneva, sans-serif" size="5" color="#f0f0ff"><b>yapCAD</b><br></font>
<font face="Verdana, Geneva, sans-serif" size="3" color="white"><b>3D drawing viewer</b></font><br>
<font face="Verdana, Geneva, sans-serif" size="1" color="white">
<br>
<b>up-arrow</b>: zoom in<br>
<b>down-arrow</b>: zoom out<br>
<b>l</b>: toggle lighting mode<br>
<b>p</b>: toggle ground plane<br>
<b>left-mouse drag</b>: rotate view<br>
<b>m</b>: toggle display of this message<br>
<b>ESC</b>: exit viewer<br>
</font>
"""

## class to provide openGL drawing functionality
class pygletDraw(drawable.Drawable):

    def window(self):
        window = []
        try:
            # Try and create a window with multisampling (antialiasing)
            config = gl.Config(sample_buffers=1, samples=4, \
                               depth_size=16, double_buffer=True)
            window = pyglet.window.Window(resizable=True, config=config)
        except pyglet.window.NoSuchConfigException:
            # Fall back to no multisampling for old hardware
            window = pyglet.window.Window(resizable=True)
        return window
    
        
    def glSetup(self):
        gl.glClearColor(0.2, 0.2, 0.3, 1)
        gl.glColor3f(1, 1, 1)
        gl.glEnable(gl.GL_DEPTH_TEST)
        gl.glEnable(gl.GL_CULL_FACE)
        gl.glEnable(gl.GL_NORMALIZE)
        
        gl.glLightfv(gl.GL_LIGHT0, gl.GL_POSITION, vec(50, 50, 10, 0))
        gl.glLightfv(gl.GL_LIGHT0, gl.GL_SPECULAR, vec(5, 5, 10, 1))
        gl.glLightfv(gl.GL_LIGHT0, gl.GL_DIFFUSE, vec(1, 1, 1, 1))
        gl.glLightfv(gl.GL_LIGHT1, gl.GL_AMBIENT, vec(0, 0, 0, 1.0))
        gl.glLightfv(gl.GL_LIGHT1, gl.GL_SPECULAR, vec(0.6, 0.6, 0.6, 1.0))
        gl.glLightfv(gl.GL_LIGHT1, gl.GL_DIFFUSE, vec(0.8, 0.8, 0.8, 1))
        gl.glLightfv(gl.GL_LIGHT1, gl.GL_POSITION, vec(-10.0, -20.0, 20.0, 0))

        # Create a Material and Group for the Model
        diffuse = [0.5, 0.5, 0.3, 1.0]
        ambient = [0.5, 0.5, 0.3, 1.0]
        specular = [1.0, 1.0, 1.0, 1.0]
        emission = [0.0, 0.0, 0.0, 1.0]
        shininess = 50
        material = pyglet.model.Material("", diffuse, ambient, specular, emission, shininess)
        self.group = pyglet.model.MaterialGroup(material=material)

        # Create a Material and Group for the ground plane
        diffuse = [0.02, 0.02, 0.023, 1.0]
        ambient = [0.01, 0.01, 0.01, 1.0]
        specular = [0.05, 0.05, 0.07, 1.0]
        emission = [0.0, 0.0, 0.0, 1.0]
        shininess = 10
        material2 = pyglet.model.Material("ground", diffuse, ambient, specular, emission, shininess)
        self.group2 = pyglet.model.MaterialGroup(material=material2)


    def makeBatches(self):
        # convert geometry lists to overall bounding box and
        # OpenGL-specific representations
        
        bbx = False
        # values for adding small offset to a point to make single-point bounding box
        epsP = point(epsilon,epsilon,epsilon)
        epsM = point(-epsilon,-epsilon,-epsilon)

        ## utility funtion to update bounding box based on a list of points
        def upbb(pp,box):
            for p in pp:
                if box:
                    box = [ [min(box[0][0],p[0]),min(box[0][1],p[1]),min(box[0][2],p[2])],
                            [max(box[1][0],p[0]),max(box[1][1],p[1]),max(box[1][2],p[2])] ]
                else:
                    box = [ add(p,epsM),add(p,epsP) ]
            return box
            
        if self.__points == [] and \
           self.__lines == [] and \
           self.__linestrips == [] and \
           self.__surfaces == []:
            raise ValueError('nothing to render')

        for p in self.__points:
            pp = p[0]
            pc = p[1]
            bbx = upbb(pp[1],bbx)
            self.__batch1.add(int(len(pp[1])/3),gl.GL_POINTS,self.group,pp,pc)

        for l in self.__lines:
            ll = l[0]
            lc = l[1]
            pp = [list(ll[1][0:3])]
            pp.append(list(ll[1][3:6]))
            bbx = upbb(pp,bbx)
            self.__batch1.add(int(len(ll[1])/3),gl.GL_LINES,self.group,ll,lc)
            
        for l in self.__linestrips:
            ll = l[0]
            lc = l[1]
            li = l[2]
            pp = []
            ln = int(len(ll[1])/3)
            for i in range(ln):
                pp.append(ll[1][i*3:(i+1)*3])
            bbx =upbb(pp,bbx)
            self.__batch1.add_indexed(int(len(ll[1])/3),
                                    gl.GL_LINES,
                                    self.group,
                                    li,ll,lc)

        for s in self.__surfaces:
            vert = s[0]
            norm = s[1]
            ind = s[2]
            pp = []
            ln = int(len(vert)/3)
            for i in range(ln):
                pp.append(vert[i*3:(i+1)*3])
            bbx = upbb(pp,bbx)
            self.__batch2.add_indexed(int(len(vert)/3),
                                    gl.GL_TRIANGLES,
                                    self.group,
                                    ind,
                                    ('v3f/static', vert),
                                    ('n3f/static', norm))

        ## Create a ground plane
        self.__batch3.add_indexed(4,
                                gl.GL_TRIANGLES,
                                self.group2,
                                [0,1,2,2,3,0],
                                ('v3f/static',[bbx[0][0]*1.2,bbx[0][1]*1.2,bbx[0][2]-0.1,
                                               bbx[1][0]*1.2,bbx[0][1]*1.2,bbx[0][2]-0.1,
                                               bbx[1][0]*1.2,bbx[1][1]*1.2,bbx[0][2]-0.1,
                                               bbx[0][0]*1.2,bbx[1][1]*1.2,bbx[0][2]-0.1]),
                                ('n3f/static',[0,0,1]*4))
                              
                                  
                                
        self.__bbox = bbx
        rnge = sub(self.__bbox[1],self.__bbox[0])
        mdim = max(rnge)
        mdim = max(mdim,10)

        # print("scene bounding box: ",vstr(bbx))

        
    def __init__(self):

        def on_key_press(symbol,modifiers):
            if symbol == pyglet.window.key.UP:
                self.__cameradist *= 1.1
                if self.__cameradist > self.__maxcameradist:
                    self.__cameradist = self.__maxcameradist
            elif symbol == pyglet.window.key.DOWN:
                self.__cameradist *= .90909090
                if self.__cameradist < self.__mincameradist:
                    self.__cameradist = self.__mincameradist
            elif symbol == pyglet.window.key.RETURN:
                self.__cameradist = self.camerastartdist
                self.__rx = 0
                self.__ry = 0
            elif symbol == pyglet.window.key.P:
                self.__drawground = not self.__drawground
            elif symbol == pyglet.window.key.M:
                self.__legend = not self.__legend
            elif symbol == pyglet.window.key.L:
                if not (self.__light0 or self.__light1):
                    self.__light0 = True
                    gl.glEnable(gl.GL_LIGHT0)
                    gl.glDisable(gl.GL_LIGHT1)
                elif self.__light0 and not self.__light1:
                    self.__light1 = True
                    gl.glEnable(gl.GL_LIGHT0)
                    gl.glEnable(gl.GL_LIGHT1)
                elif self.__light0 and self.__light1:
                    self.__light0 = self.__light1 = False
            
                    
        self.__window = self.window()
        self.__window.push_handlers(on_key_press)
        self.__window.projection = pyglet.window.Projection3D()
        self.glSetup()
        self.__center= point(0,0)
        self.__magnify = 0.08
        self.__arcres = 5
        self.__cameradist = self.camerastartdist = 100.0
        self.__maxcameradist = 255.0
        self.__mincameradist = 10.0
        self.__light0 = False
        self.__light1 = False
        self.__legend = True
        self.__drawground = True
        self.__rx = 0
        self.__ry = 0
        self.__points= []
        self.__lines= []
        self.__linestrips=[]
        self.__surfaces=[]
        self.__labels = []
        self.__batch1 = graphics.Batch() # use for lines, points, etc.
        self.__batch2 = graphics.Batch() # use for surfaces
        self.__batch3 = graphics.Batch() # use for environmental features, e.g. ground plane
        self.__bbox = False

        @self.__window.event
        def on_mouse_drag(x, y, dx, dy, buttons, modifiers):
            self.__rx += dx
            self.__ry -= dy
            return pyglet.event.EVENT_HANDLED



        @self.__window.event
        def on_draw():
            self.__window.clear()
            gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
            
            self.__window.projection = pyglet.window.Projection3D()
            # gl.glEnable(gl.GL_BLEND)
            gl.glColor3f(1., 1., 1.)

            #gl.glMatrixMode(gl.GL_MODELVIEW)
            gl.glLoadIdentity()
            cent = scale3(add(self.__bbox[0],self.__bbox[1]),0.5)
            rnge = sub(self.__bbox[1],self.__bbox[0])
            mdim = max(rnge)
            mdim = max(mdim,10)
            
            gl.glTranslatef(-cent[0],-cent[1],-cent[2]-1*self.__cameradist)
            #gl.glTranslatef(0,0,-1*self.__cameradist)
            gl.glRotatef(self.__ry%360.0, 1, 0, 0)
            gl.glRotatef(self.__rx%360.0, 0, 1, 0)

            #draw "ground plane"
            if self.__drawground:
                gl.glEnable(gl.GL_LIGHTING)
                gl.glEnable(gl.GL_LIGHT0)
                self.__batch3.draw()  

            gl.glDisable(gl.GL_LIGHTING)
            self.__batch1.draw()
            
            if self.__light0 or self.__light1:
                gl.glEnable(gl.GL_LIGHTING)
                self.__batch2.draw()

            #gl.glDisable(gl.GL_BLEND)
            gl.glDisable(gl.GL_LIGHTING)
            gl.glColor3f(1., 1., 1.)
            gl.glScalef(0.05, 0.05, 0.05)
            for label in self.__labels:
                label.draw()
                
            if self.__legend:
                gl.glMatrixMode(gl.GL_MODELVIEW)
                gl.glPushMatrix()
                gl.glLoadIdentity()
                gl.glMatrixMode(gl.GL_PROJECTION)
                gl.glPushMatrix()
                gl.glLoadIdentity()

                self.__window.projection = pyglet.window.Projection2D()
                label = pyglet.text.HTMLLabel(yapCAD_legend,
                                              x=10, y=30,
                                              anchor_x='left',anchor_y='bottom',
                                              width=400,
                                              multiline=True)
                label.draw()
                gl.glMatrixMode(gl.GL_PROJECTION)
                gl.glPopMatrix()
                gl.glMatrixMode(gl.GL_MODELVIEW)
                gl.glPopMatrix()
         
            return pyglet.event.EVENT_HANDLED
            
        super().__init__()

    def __repr__(self):
        return 'an instance of pygletDraw'

    ## properties

    @property
    def magnify(self):
        return self.__magnify

    def _set_magnify(self,mag):
        self.__magnify=mag

    @magnify.setter
    def magnify(self,mag=False):
        if not isinstance(mag,(int,float)):
            raise ValueError('invalid magnification ' + str(mag))
        if isinstance(mag,bool) and mag ==False:
            mag = 1.0
        elif mag < epsilon:
            mag = epsilon
        self._set_magnify(mag)

    @property
    def cameradist(self):
        return self.__cameradist

    def _set_cameradist(self,dist):
        self.__cameradist=dist

    @cameradist.setter
    def cameradist(self,dist=False):
        if not isinstance(dist,(int,float)):
            raise ValueError('invalid camera distance ' + str(dist))
        if isinstance(dist,bool) and dist ==False:
            dist = 100.0
        elif dist < self.__mincameradist:
            dist = self.__mincameradist
        self._set_cameradist(dist)

    
    ## Overload virtual yapcad.drawable base class dawing methods
    
    def draw_point(self,p):
        ar = self.__arcres
        self.__arcres=30
        super().draw_point(p)
        self.__arcres=ar

    def draw_line(self,p1,p2):
        color = self.thing2color(self.linecolor,'f')
        self.__lines.append([ ('v3f',(p1[0],p1[1],p1[2],
                                      p2[0],p2[1],p2[2])),
                              ('c3f',tuple(color + color)) ])
            
    def draw_arc(self,p,r,start,end):
        res = self.__arcres
        # resolution of arc sampling in degrees
        points = []
        if not (start==0 and end==360):
            start = start%360.0
            end = end % 360.0
            if end < start:
                end = end + 360
        theta = start*pi2/360.0
        points.append(add(p,[math.cos(theta)*r,math.sin(theta)*r,0.0,1.0]))
        for a in range(round(start),round(end),res):
            theta = a*pi2/360.0
            pp = [math.cos(theta)*r,math.sin(theta)*r,0.0,1.0]
            pp = add(pp,p)
            points.append(pp)
        theta = end*pi2/360.0
        points.append(add(p,[math.cos(theta)*r,math.sin(theta)*r,0.0,1.0]))

        self.draw_linestrip(points)

    def draw_text(self,text,location,
                  align='left',
                  attr={'name': 'Times New Roman',
                        'size': 12}):
        name = 'Times New Roman'
        if 'font_name' in attr:
            name = attr['font_name']
        bold = False
        if 'bold' in attr:
            bold = attr['bold']
        italic = False
        if 'italic' in attr:
            italic = attr['italic']
        underline = False
        if 'underline' in attr:
            underline = attr['underline']
        size = 12
        if 'size' in attr:
            size = attr['size']
        elif 'height' in attr:
            size *= attr['height']
        anchor_x='left'
        if 'anchor_x' in attr:
            anchor_x = attr['anchor_x']
        anchor_y='center'
        if 'anchor_y' in attr:
            anchor_y = attr['anchor_y']
        col = []
        if 'color' in attr:
            col = attr['color']
        elif self.linecolor:
            col = self.thing2color(self.linecolor,'b')
            print ("color: ",col)
        else:
            col = [255,255,255]
            
        color = tuple(col + [255])
            
        x = location[0]
        y = location[1]
        
        self.__labels.append(pyglet.text.Label(text,
                                               font_name=name,
                                               font_size=size*self.__magnify,
                                               x=x*20.0, #fudge factor to make font rendering look OK
                                               y=y*20.0,
                                               align=align,
                                               color=color,
                                               bold=bold,
                                               italic=italic,
                                               #underline=underline,
                                               anchor_x=anchor_x,
                                               anchor_y=anchor_y))
        
    ## OpenGL-specific drawing methods
    
    def draw_linestrip(self,points):
        # we simulate a linestrip useing GL_LINES and indexed drawing.
        # This prevents extra lines
        p2 = []
        i2 = []
        color = self.thing2color(self.linecolor,'f')
        for p in points:
            p2 = p2 + p[0:3]
        for i in range(1,len(points)):
            i2 = i2 + [i-1, i]
        c2 = color * int(len(p2)/3)
        self.__linestrips.append([ ('v3f', tuple(p2)),\
                                   ('c3f',tuple(c2)),
                                   tuple(i2) ])
            

    def draw_surface(self,points,normals,faces):
        vrts = []
        nrms= []
        inds = []
        for p in points:
            vrts+=p[0:3]
        for n in normals:
            nrms+=n[0:3]
        for f in faces:
            inds+=f[0:3]
        self.__surfaces.append([vrts,nrms,inds])    

    ## overload base-class virtual display method
    def display(self):
        self.makeBatches()
        pyglet.app.run()




import pytest
from geom import *
## unit tests for yapCAD geom.py

class TestPoint:
    """unit tests for yapCAD point functions"""

    def test_create(self):
        a = point(5,0)
        b = point(0,5,-2)
        c = point(-2.3,4.6,-9.2,0.5)
        bb = point(b)
        assert a == [5,0,0,1]
        assert b == [0,5,-2,1]
        assert c == [-2.3,4.6,-9.2,0.5]
        assert bb[0] == b[0] and bb[1] == b[1] \
            and bb[2] == b[2] and bb[3] == b[3]

    def test_discrimate(self):
        a = point(5,0)
        b = point(0,5)
        assert ispoint(a)
        assert ispoint(b)
        assert not ispoint(vect(1,2,3,-1))
        assert ispoint([0,2,2,1])
        assert not ispoint([1,2])

    def test_format(self):
        a = point(5,0)
        b = point(2,3,2)
        c = point(1,2,3,4)
        assert vstr(a) == '[5, 0]'
        assert vstr(b) == '[2, 3, 2]'
        assert vstr(c) == '[1, 2, 3, 4]'

class TestOperations:
    def test_vect(self):
        
        a = point(5,0)
        b = point(0,5)
        c = point(-3,-3)
        d = point(1,1)
        e = point(10,10)
        assert close(mag(a),5.0)
        assert vclose(add(a,b),point(5,5))
        assert vclose(sub(a,b),point(5,-5))
        assert close(dot(a,b),0)
        assert close(dot(d,c),-6)
        assert vclose(cross(a,b),point(0,0,25))
        assert vclose(cross(b,a),point(0,0,-25))
        assert close(mag(sub(a,b)),sqrt(50))

class TestLine:
    def test_create(self):
        
        a = point(5,0)
        b = point(0,5)
        c = point(-3,-3)
        d = point(1,1)
        e = point(10,10)
        f = vect(1,2,3,-1)

        assert line(a,b) == [a,b]
        assert isline(line(a,b))
        assert isline([a,b])
        assert not isline([a,f])
        with pytest.raises(ValueError):
            assert not lsine(line(a,f))
        
    def test_sample(self):
        a = point(-5,-1)
        b = point(5,3)
        l = line(a,b)
        print("sample tests")
        p0 =sampleline(l,-0.5)
        p1 =sampleline(l,0.0)
        p2 =sampleline(l,0.5)
        p3 =sampleline(l,1.0)
        p4 =sampleline(l,1.5)
        assert vclose(p0,point(-10,-3))
        assert vclose(p1,a)
        assert vclose(p2,point(0,1))
        assert vclose(p3,b)
        assert vclose(p4,point(10,5))
        print("unsample tests")
        u0 = unsampleline(l,p0)
        u1 = unsampleline(l,p1)
        u2 = unsampleline(l,p2)
        u3 = unsampleline(l,p3)
        u4 = unsampleline(l,p4)
        assert close(u0,-0.5)
        assert close(u1,0.0)
        assert close(u2,0.5)
        assert close(u3,1.0)
        assert close(u4,1.5)
        p = point(100,100)
        assert not unsampleline(l,p)

    def test_intersect(self):
        print("---> line-line intersection tests")
        a = point(5,0)
        b = point(0,5)
        c = point(-3,-3)
        d = point(1,1)
        e = point(10,10)
        l1 = line(a,b)
        print("l1:",vstr(l1))
        l2 = line(c,d)
        l3 = line(c,e)
        print("l1:" + vstr(l1) + ", l2:" + vstr(l2) +", l3:" + vstr(l3))

        int0 = lineLineIntersectXY(l1,l1)
        int0u = lineLineIntersectXY(l1,l1,params=True)
        int1 = lineLineIntersectXY(l1,l2,False)
        int1u = lineLineIntersectXY(l1,l2,params=True)
        int2 = lineLineIntersectXY(l1,l2,True)
        int2u = lineLineIntersectXY(l1,l2,params=True)
        int3 = lineLineIntersectXY(l1,l3,True)
        int3u = lineLineIntersectXY(l1,l3,params=True)
        
        assert not int0
        assert not int0u
        assert vclose(int1,point(2.5,2.5))
        assert len(int1u) == 2
        assert close(int1u[0],0.5)
        assert close(int1u[1],1.375)
        assert not int2
        assert close(int2u[0],0.5)
        assert close(int2u[1],1.375)
        assert vclose(int3,point(2.5,2.5))
        assert close(int3u[0], 0.5)
        assert close(int3u[1], 0.4230769230769231)
        
class TestArc:
    def test_create(self):
        arc1=[vect(2.5,2.5),vect(2.5,90.0,270.0,-1)]
        arc11=arc(point(2.5,2.5),2.5,90.0,270.0)
        assert isarc(arc1)
        assert isarc(arc11)
        assert arc1 == arc11
        arc12=arc(arc11)
        assert isarc(arc12)
        assert arc12==arc11
        arc11[1][0]=5.0         # modify arc11
        assert arc12[1][0]==2.5 # assert that this didn't change arc12
        with pytest.raises(ValueError):
            assert not arc(vect(0,0),-2)

    def test_sample(self):
        arc1=arc(point(0,0),1.0,90.0,270.0)
        print("sample tests")
        print("arc1: ",vstr(arc1))
        for i in range(5):
            u = i/4
            theta = 90+i*45.0
            p = samplearc(arc1,u)
            x = point(cos(theta*pi2/360.0),sin(theta*pi2/360.0))
            print("sample: ",i," theta: ",theta)
            print("  sampled point p: ",vstr(p))
            print("  reference point x: ",vstr(x))
            print("  assert that these are the same")
            assert vclose(p,x)
        print("unsample tests")
        for i in range(5):
            u = i/4
            theta = 90+i*45.0
            x = point(cos(theta*pi2/360.0),sin(theta*pi2/360.0))
            uu = unsamplearc(arc1,x)
            print("unsample: ",i," theta: ",theta)
            print("  test point x: ",vstr(x))
            print("  unsampled parameter uu: ",uu)
            print("  calculated parameter u: ",u)
            print("  assert that these are the same")
            assert close(u,uu)

    def test_intersect_line(self):
        print("---> line-arc intersection testing")
        arc1=[vect(2.5,2.5),vect(2.5,90.0,270.0)]
        print("arc1: {}".format(vstr(arc1)))
        a = point(5,0)
        b = point(0,5)
        c = point(-3,-3)
        d = point(0.0)
        l1 = line(a,b)
        print("l1: ", vstr(l1))
        l2 = line(c,d)
        print("l2: ",vstr(l2))
        int4 = lineArcIntersectXY(l1,arc1,False)
        int4u = lineArcIntersectXY(l1,arc1,params=True)
        print("cross-verifying intersection points and samples")
        for i in range(2):
            u1 = int4u[0][i]
            u2 = int4u[1][i]
            p=int4[i]
            p1 = sampleline(l1,u1)
            p2 = samplearc(arc1,u2)
            uu1 = unsampleline(l1,p)
            uu2 = unsamplearc(arc1,p)
            print("i: ",i,"u1: ",u1,"u2: ", u2,"p: ",vstr(p),
                  "p1: ",vstr(p1),"p2: ",vstr(p2))
            print("uu1: ",uu1," uu2: ",uu2)
            assert close(uu1,u1)
            assert close(uu2,u2)
            assert vclose(p,p1)
            assert vclose(p,p2)

        int5 = lineArcIntersectXY(l1,arc1,True)
        int6 = lineArcIntersectXY([vect(0,5),vect(5,5)],arc1,True)
        int7 = lineArcIntersectXY(l2,arc1,True)
        int8 = lineArcIntersectXY(l2,arc1,False)

        p1 = point(4.267766952966369, 0.7322330470336311)
        p2 = point(0.7322330470336311, 4.267766952966369)
        print("lineArcIntersectXY(l1,arc1,False): ", vstr(int4))
        assert len(int4) == 2
        assert vclose(int4[0],p1)
        assert vclose(int4[1],p2)
        print("lineArcIntersectXY(l1,arc1,params=True)",vstr(int4u))
        print("lineArcIntersectXY(l1,arc1,True): ", vstr(int5))
        assert len(int5) == 1
        assert vclose(int5[0],p2)
        print("tangent line, one intersection test")
        assert len(int6) == 1
        print("lineArcIntersectXY([vect(0,5),vect(5,5)],arc1,True): ",
              vstr(int6))
        assert vclose(int6[0],point(2.5,5.0))
        print("lineArcIntersectXY(l2,arc1,False): ",vstr(int7))
        assert not int7
        print("lineArcIntersectXY(l2,arc1,True): ",vstr(int8))
        assert len(int8) == 2
        assert vclose(int8[0], point(0.7322330470336311, 0.7322330470336311))
        assert vclose(int8[1], point(4.267766952966369, 4.267766952966369))


    def test_intersect_arc(self):
        print("---> arc-arc intersection tests")
        arc1=arc(vect(-2.5,0),2.5,90.0,270.0)
        arc2=arc(vect(2.5,0),2.5,90.0,270.0)

        print("arc1: {}".format(vstr(arc1)))
        print("arc2: {}".format(vstr(arc2)))
        print("single point of intersection test")
        int9 = arcArcIntersectXY(arc1,arc2,False)
        int9u =  arcArcIntersectXY(arc1,arc2,params=True)
        print("arcArcIntersectXY(arc1,arc2,False):",vstr(int9))
        print("arcArcIntersectXY(arc1,arc2,params=True):",int9u)
        
        assert len(int9) == 1
        assert vclose(int9[0],point(0,0))
        assert len(int9u) == 2
        assert close(int9u[0][0],-0.5)
        assert close(int9u[1][0],0.5)
        int10 = arcArcIntersectXY(arc1,arc2,True)
        int11 = arcArcIntersectXY(arc2,arc1,True)
        
        print("arcArcIntersectXY(arc1,arc2,True):",vstr(int10))
        print("arcArcIntersectXY(arc2,arc1,True):",vstr(int11))
        assert not int10
        assert not int11

        x = cos(pi/4)
        arc3=arc(vect(-x,-x),2.0,315.0,135.0)
        arc4=arc(vect(x,x),2.0,135.0,315.0)
        print("arc3: {}".format(vstr(arc3)))
        print("arc4: {}".format(vstr(arc4)))
        int12 = arcArcIntersectXY(arc3,arc4,True)
        int12u = arcArcIntersectXY(arc3,arc4,params=True)
        int13 = arcArcIntersectXY(arc3,arc4,False)
        print("arcArcIntersectXY(arc3,arc4,True):",vstr(int12))
        print("arcArcIntersectXY(arc3,arc4,False):",vstr(int13))
        print("arcArcIntersectXY(arc3,arc4,params=True):",int12u)
        assert len(int12u) == 2
        assert len(int12u[0])==2
        assert len(int12u[1])==2
        print("cross-verifying intersection points and samples")
        for i in range(2):
            u1 = int12u[0][i]
            u2 = int12u[1][i]
            p=int12[i]
            p1 = samplearc(arc3,u1)
            p2 = samplearc(arc4,u2)
            uu1 = unsamplearc(arc3,p)
            uu2 = unsamplearc(arc4,p)
            print("i: ",i,"u1: ",u1,"u2: ", u2,"p: ",vstr(p),
                  "p1: ",vstr(p1),"p2: ",vstr(p2))
            print("uu1: ",uu1," uu2: ",uu2)
            assert close(uu1,u1)
            assert close(uu2,u2)
            assert vclose(p,p1)
            assert vclose(p,p2)
            
        
class TestUtility:
    def test_copy(self):
        a = point(-5,-1)
        b = point(5,3)
        c = point(-3,-3)
        d = point(1,1)
        e = point(10,10)
        l = line(a,b)
        l2 = [c,d]
        l3 = line(c,e)
        foo = [a,b,l3,l2,d]
        foobar = deepcopy(foo)
        assert foobar == foo
        foo[0]=e
        assert not foobar==foo
        bar = [a,b,[1,2],l2,l3]
        assert not deepcopy(bar)

class TestPoly:
    def test_create(self):
        with pytest.raises(ValueError):
            poly()

        a = point(0,-5)
        b = point(5,0)
        c = point(0,5)
        d = point(-5,0)

        pol1 = poly(a,b,c)
        assert pol1 == [a,b,c]

        pol2 = poly(a,b,c,d)
        assert pol2 == [a,b,c,d]
        
        pol3 = poly(pol1)
        assert not pol2 == pol1
        assert pol3 == pol1

        with pytest.raises(ValueError):
            poly(a,b,'foo')

        with pytest.raises(ValueError):
            poly(a,b)

    def test_discriminate(self):
        a = point(0,-5)
        b = point(5,0)
        c = point(0,5)
        
        pol1 = poly(a,b,c)
        pol2 = poly(a,b,c,a)
        
        assert ispoly(pol1)
        assert not ispoly(a)
        assert not ispoly('foo')

        assert ispoly(pol2)
        assert ispolygon(pol2)
        assert not ispolygon(pol1)
        assert not ispolygon([])

    def test_sample(self):
        a = point(0,-5)
        b = point(5,0)
        c = point(0,5)
        d = point(-5,0)

        pol1 = poly(a,b,c,d)
        print("---> tests for sampling and unsampling of polylines")
        print("pol1: ",vstr(pol1))
        pp = []
        print("sampled points: ")
        for i in range(9):
            u = i/8
            p = samplepoly(pol1,u)
            pp = pp+ [ p ]
            print("i: ",i,"u: ",u," p: ",vstr(p))
        print("unsampling polygon")
        for i in range(len(pp)):
            u = i/(len(pp)-1)
            uu = unsamplepoly(pol1,pp[i])
            print("i: ",i," p: ",vstr(pp[i])," u: ",u," uu: ",uu)
            assert close(u,uu)
            
    def test_intersect(self):
        a = point(0,-5)
        b = point(5,0)
        c = point(0,5)
        d = point(-5,0)
        
        pol1 = poly(a,b,c,d)
        pol2 = poly(a,b,c,d,a)

        line1 = line(point(0,0),point(5,5))
        line2 = line(point(0,-10),point(0,10))

        arc1 = arc(point(0,0),4.0,270,90)
        arc2 = arc(point(0,0),5.0,0,360)

        print("--> polyline and polygon intersection testing")
        print("pol1: ",vstr(pol1))
        print("pol2: ",vstr(pol2))
        print("line1: ",vstr(line1))
        print("line2: ",vstr(line2))
        print("arc1: ",vstr(arc1))
        print("arc2: ",vstr(arc2))
        
        int0 = intersectSimplePolyXY(line1,pol1,True)
        int1 = intersectSimplePolyXY(line1,pol1,False)
        int0u = intersectSimplePolyXY(line1,pol1,params=True)
        print("intersectSimplePolyXY(line1,pol1,True): ",vstr(int0))
        print("intersectSimplePolyXY(line1,pol1,False): ",vstr(int1))
        print("intersectSimplePolyXY(line1,pol1,params=True): ",vstr(int0u))
        assert len(int0) == 1
        assert vclose(int0[0],point(2.5,2.5))
        assert len(int1) == 1
        assert vclose(int1[0],point(2.5,2.5))
        assert len(int0u[0]) == 1
        assert close(int0u[0][0],0.5)
        assert close(int0u[1][0],0.5)

        int0 = intersectSimplePolyXY(line1,pol2,True)
        int1 = intersectSimplePolyXY(line1,pol2,False)
        int0u = intersectSimplePolyXY(line1,pol2,params=True)
        print("intersectSimplePolyXY(line1,pol2,True): ",vstr(int0))
        print("intersectSimplePolyXY(line1,pol2,False): ",vstr(int1))
        print("intersectSimplePolyXY(line1,pol2,params=True): ",vstr(int0u))
        assert len(int0) == 1
        assert vclose(int0[0],point(2.5,2.5))
        assert len(int1) == 2
        assert vclose(int1[0],point(2.5,2.5))
        assert vclose(int1[1],point(-2.5,-2.5))
        assert len(int0u[0]) == 2
        assert close(int0u[0][0],0.375)
        assert close(int0u[0][1],0.875)
        assert close(int0u[1][0],0.5)
        assert close(int0u[1][1],-0.5)

        int0 = intersectSimplePolyXY(arc1,pol2,True)
        int1 = intersectSimplePolyXY(arc1,pol2,False)
        int0u = intersectSimplePolyXY(arc1,pol2,params=True)
        print("intersectSimplePolyXY(arc1,pol2,True): ",vstr(int0))
        print("intersectSimplePolyXY(arc1,pol2,False): ",vstr(int1))
        print("intersectSimplePolyXY(arc1,pol2,params=True): ",vstr(int0u))


        assert False



# Polyline and Polygon geometry generating classes for yapCAD
from yapcad.geom import *
from yapcad.geometry import *

## These geometry-generating classes are somewhat analogous to the
## polygonal representations available by way of poly() point lists.
## There are two important diferences that might lead you to use these
## classes rather than the point-list based representations.

## First, because instances of these classes can cache
## meta-information about the geometry, it means that sampling and
## intersection operations might be more efficient for class instances
## than utilizing the functions avalable for the list-based
## representations. 

## Seond, and more importantly, these geometry-generating classes,
## they can support a richer set of representations and operations.
## For example, the Polygon() class for closed figures allows you to
## specify points, lines, arcs, or circles as elements of the
## geometry, and will intelligently interpolate between these to
## generate the resulting geometry.  This means you can use circles to
## represented rounded corners without having to worry about
## converting them into just the right arc, since the tangent lines
## and arc trimming are done automatically.


class Polyline(IntersectGeometry):
    """Generalized multi-element open figure class"""

    def __init__(self,a=False):
        self._elem=[]
        self._length=0.0
        self._lengths=[]
        self._lines=[]
        self._update=True
        self._closed=False
        self._center=point(0,0,0)
        self._bbox=line(point(-epsilon,-epsilon),
                        point(epsilon,epsilon))
        if isinstance(a,Polyline):
            self._elem = deepcopy(a._elem)
            self._updateInternals()
        elif isgeomlist(a):
            for i in a:
                if ispoint(i) or isarc(i) or isline(i): 
                    self._elem.append(deepcopy(i))
                elif ispoly(i):
                    self._elem = self._elem + i
                else:
                    raise ValueError("bad argument to Polyline constructor")
        elif a != False:
            raise ValueError("bad argument passed to Polyline constructor")

    def __repr__(self):
        return 'Polyline({})'.format(vstr(self._elem))

    ## add another drawing element
    def addPoint(self,element):
        if ispoint(element):
            self._update=True  # flag that we need to recalculate stuff
            self._elem.append(element)
        else:
            raise ValueError('attempt to add a non point to Polyline')

    ## set a point to a specified value
    def setPoint(self,i,p):
        if not ispoint(p):
            raise ValueError('bad point passed to Polyline.setPoint(): '.format(p))
        if i < len(self._elem):
            self._elem[i]=point(p)
        elif i == len(self._elem):
            self._elem.append(p)
        else:
            raise ValueError('index out of range in PolylinesetPoint(): '.format(i))
        self._update=True
        
    ## return a copy of the elem list
    def getElem(self):
        return deepcopy(self._elem)
    
    def _updateInternals(self):
        if self._update:
            self._updateCenter()
            self._updateLines()
            self._update=False


    ## compute barycentric center of figure by equally weighting the
    ## points.  If last and first points are the same, ignore the
    ## last point.
    def _updateCenter(self):
        l = len(self._elem)
        if l == 0:
            return
        elif l == 1:
            self._center = center(self._elem[0]) # center is sole point
            self._bbox = bbox(self._elem[0])
        else:
            
            if dist(center(self._elem[0]),
                    center(self._elem[-1])) < epsilon:
                l -= 1

            
            p = center(self._elem[0])
            for i in range(1,l):
                p = add(center(self._elem[i]),p)

            self._center = scale3(p,1/l)
            self._bbox = bbox(self._elem)
        return

    ## return the center of the figure.  If necessary, recompute that
    def getCenter(self):
        if self._update:
            self._updateInternals()
        return self._center


    def getLength(self):
        if self._update:
            self._updateInternals()
        return self._length
    
    ## function to take the points in the elem[] list and build a
    ## list of the individual lines, line lengths and total length to
    ## facilitate sampling
    
    def _updateLines(self):
        for i in range(1,len(self._elem)):
            p0=  center(self._elem[i-1])
            p1=  center(self._elem[i])
            self._lines.append(line(p0,p1))
            l = dist(p0,p1)
            self._lengths.append(l)
            self._length += l
        if len(self._elem) > 2 and dist(center(self._elem[0]),
                                        center(self._elem[-1])) < epsilon:
            self._closed = True

    def geom(self):
        if self._update:
            self._updateInternals()
        return deepcopy(self._lines)

    def sample(self,u):
        if self._update:
            self._updateInternals()
        dist = u * self._length;
        d = 0
        if self._closed:
            u = u%1.0
        if u < 1.0:
            for i in range(len(self._lines)):
                l=self._lengths[i]
                if dist <= d+l:
                    uu = 1.0 - (d+l-dist)/l
                    return sampleline(self._lines[i],uu)
                else:
                    d+=l
        else:
            uu = (dist-self._length+self._lengths[-1])/self._lengths[-1]
            return sampleline(self._lines[-1],uu)
        

class Polygon(Polyline):
    """Generalized multi-element closed figure geometry generation class"""

    def __init__(self,a=False):
        super().__init__()
        if isinstance(a,Polygon):
            self._elem = deepcopy(a._elem)
            self._updateInternals()
        elif isgeomlist(a):
            for i in a:
                if ispoint(i) or isarc(i):
                    self._elem.append(i)
                elif isline(i):
                    pass
                elif ispoly(i):
                    self._elem = self._elem + i
                else:
                    raise ValueError("bad argument to Polygon constructor")

        elif a != False:
            raise ValueError("bad argument to Polygon constructor")

        elm = self._elem

        l = len(elm)
        if l > 2:
            # if first and last element are identicial points or circles,
            # remove them
            if (ispoint(elm[0]) and ispoint(elm[-1]) and \
                dist(elm[0],elm[-1]) < epsilon) or \
               (iscircle(elm[0]) and iscircle(elm[-1]) and \
               abs(elm[0][1][0] - elm[-1][1][0]) < epsilon):
                self._elem.pop()
        
    def __repr__(self):
        return 'Polygon({})'.format(vstr(self._elem))

    def _updateInternals(self):
        if self._update:
            self._updateCenter()
            self._makeoutline()
            self._update=False

    ## add another drawing element
    def addLine(self,element):
        if isline(element):
            self._update=True  # flag that we need to recalculate stuff
            self._elem.append(deepcopy(element))
        else:
            raise ValueError('attempt to add a non point, line or arc to poly')
    
    ## add another drawing element
    def addArc(self,element):
        if isarc(element):
            self._update=True  # flag that we need to recalculate stuff
            self._elem.append(deepcopy(element))
        else:
            raise ValueError('attempt to add a non point, line or arc to poly')
                
    def remove(self,element):
        self._update=True  # flag that we need to recalculate stuff
        self._elem.remove(element)
                
    ## function to take the elements in the elem[] list and
    ## construct the full outline.  For example, consider a list of
    ## three points in the elem[] list -- this would result in the
    ## construction of three lines that results in a closed curve.
    ## Similarly, three circles in the elem list would result in a
    ## "rounded triangle" composed of the circle-tangent lines joined
    ## by three arcs.  Elements that are explicitly specified lines or
    ## non-circular arcs are joined to adjacent elements by lines

    def _makeoutline(self):
        def _calclength():
            l = 0
            ll = []
            length = 0
            for o in self._outline:
                if isarc(o):
                    length=arclength(o)
                elif isline(o): # line
                    length=linelength(o)
                else:
                    #print("self._outline: ",vstr(self._outline))
                    #print("o: ",vstr(o))
                    raise ValueError("bad element in outline list for _calclength()")
                l += length
                ll.append(length)
            self._length = l
            self._lengths=ll

        def _handleCircle(e0,e1,e2):
            ## get the two tangent lines from circle e1 to the circle
            ## e2
            ll = circleCircleTangentsXY(e1,e2)
            l=[]
            x0=center(e0)
            x1=center(e1)
            x2=center(e2)
            #print("x0: ",vstr(x0)," x1: ",vstr(x1)," x2: ",vstr(x2))
            #print("ll: ",vstr(ll))
            v1= sub(x1,x0)
            v2= sub(x2,x1)
            r0 = cross(v1,v2)
            x3 = linecenter(ll[0])
            x4 = linecenter(ll[1])
            #print("x3: ",vstr(x3),"x4: ",vstr(x4))
            #print("v1: ",vstr(v1),"v2: ",vstr(v2))
            v3 = sub(x3,x2)
            v4 = sub(x4,x2)
            r1 = cross(v2,v3)
            r2 = cross(v2,v4)
            #print("v3: ",vstr(v3)," v4: ",vstr(v4))
            #print("r0: ", r0[2], " r1: ",r1[2]," r2: ",r2[2])

            if r0[2] >= 0:
                #print("left-hand turn")
                if r1[2] >= 0:
                    l = ll[1]
                else:
                    l = ll[0]
            else:
                #print("right hand turn")
                if r2[2] >= 0:
                    l = ll[0]
                else:
                    l = ll[1]
                        
            self._outline.append(line(l))
            
        def _fromPointAdd(e0,p1,e2):
            if ispoint(e2): #r1 is a point -- simplest case
                #ll = dist(p1,e2)
                self._outline.append(line(p1,e2))
            elif isline(e2):
                self._outline.append(line(p1,e2[0]))
                self._outline.append(line(e2))
            elif isarc(e2) and not iscircle(e2):
                p = samplearc(e2,1*epsilon)
                self._outline.append(line(p1,p))
                self._outline.append(arc(e2))
            elif iscircle(e2): 
                _handleCircle(e0,arc(p1,1*epsilon),e2)
                self._outline.append(arc(e2))
            else:
                raise ValueError('bad object in element list')
            
        self._outline=[]
        self._length=0
        self._lengths=[]

        if len(self._elem) < 2:
            ## if one element, the outline is the element
            if len(self._elem) == 1:
                self._outline=deepcopy(self._elem)
                _calclength()
            ## if there are fewer than one elem, there is nothing to do
            return
        ## we have work to do.  Construct all lines and circle-tangent
        ## lines first, inserting explicit line and arc elements.
        ## Then go back and convert circles to arcs.  Finally compute
        ## total length and length list for use in sample() function.

        
        for i in range(len(self._elem)):
            if i == 0:
                e0 = self._elem[-1]
            else:
                e0 = self._elem[i-1]
            e1 = self._elem[i]
            if i == len(self._elem)-1: #last item
                e2 = self._elem[0]
            else:
                e2 = self._elem[i+1]
            ## work through element types
            if ispoint(e1): #e1 is a point
                _fromPointAdd(e0,e1,e2)
            elif isline(e1):
                _fromPointAdd(e0,e1[0],e2)
            elif isarc(e1) and not iscircle(e1):
                p1 = samplearc(e1,1.0)
                _fromPointAdd(e0,p1,e2)
            elif iscircle(e1):
                c2 = []
                if ispoint(e2):
                    c2 = arc(e2,1*epsilon)
                if iscircle(e2):
                    c2 = e2
                else:
                    p = sample(e2,0.0)
                    c2 = arc(p,1*epsilon)
                _handleCircle(e0,e1,c2)
                
                if not ispoint(e2):
                    self._outline.append(deepcopy(e2))

        ## OK, now go back through and replace full circles with arcs
        ## and catch any intersecting lines due to non-convex
        ## curvature
        for i in range(1,len(self._outline)):
            e0 = self._outline[i-1]
            e1 = self._outline[i]
            if i == len(self._outline)-1: #last item
                e2 = self._outline[0]
            else:
                e2 = self._outline[i+1]
            if isline(e1) and isline(e2):
                pi = lineLineIntersectXY(e1,e2,inside=True)
                if pi == False:
                    # print("not expected: adjacent lines don't intersect. We will fix that")
                    pi = lineLineIntersectXY(e1,e2,inside=False)
                    if pi == False:
                        # raise ValueError('parallel adjacent lines -- no clue')
                        print("odd parallel? adjacent line condition")
                        continue
                    else:
                        self._outline[i] = line(e1[0],pi)
                        self._outline[(i+1)%len(self._outline)] = line(pi,e2[1])
            if iscircle(e1):
                if not isline(e0) or not isline(e2):
                    raise ValueError('circle not bracketed by lines')
                pp0 = lineArcIntersectXY(e0,e1,False)
                pp1 = lineArcIntersectXY(e2,e1,False)
                # these should be tangent lines, so exactly one
                # intersection each.  sometimes we get two that are
                # just a bit more than epsilon apart, which is OK.
                if len(pp0) < 1 or len(pp1) < 1:
                    raise ValueError('bad line-circle intersection in poly outline calculation')
                # we are assuming that elements are ordered in a
                # counter-clockwise fashion
                p0=sub(pp0[0],e1[0])
                p1=sub(pp1[0],e1[0])
                start = (atan2(p0[1],p0[0]) % pi2) * 360.0/pi2
                end = (atan2(p1[1],p1[0]) % pi2) * 360.0/pi2
                newarc = arc(e1[0],e1[1][0],start,end)
                self._outline[i]=newarc

        self._outline = cullZeroLength(self._outline)
                
        _calclength()
        self._bbox = geomlistbbox(self._outline)
        
    def sample(self,u):
        if self._update:
            self._updateInternals()
        if len(self._outline) == 0:
            raise ValueError('no geometry to sample, empty poly')
        dist = (u % 1.0) * self._length
        d = 0
        for i in range(len(self._outline)):
            l=self._lengths[i]
            if dist <= d+l:
                u = 1.0 - (d+l-dist)/l
                e = self._outline[i]
                return sample(e,u)
            else:
                d=d+l

    def segment(self,u1,u2,reverse=False):
        return segmentgeomlist(self.geom(),u1,u2,closed=True,reverse=reverse)

    def mirror(self,plane,poly=False):
        if poly:
            p = Polygon()
            p._elem = deepcopy(self._elem)
            p._elem.reverse()
            p._elem = mirror(p._elem,plane)
            p._update=True
            return p
        
        return mirror(self.geom(),plane)

    def rotate(self,angle,cent=point(0,0,0),axis=point(0,0,1),poly=False):
        if poly:
            p = Polygon(self)
            p._elem = rotate(self._elem,angle,cent,axis)
            p._update = True
            return p
        
        return rotate(self.geom(),angle,cent,axis)

    def scale(self,sx,sy=False,sz=False,cent=point(0,0),poly=False):
        if poly:
            p = Polygon(self)
            p._elem = scale(self._elem,sx,sy,sz,cent)
            p._update = True
            return p

        return scale(self.geom(),sx,sy,sz,cent)

    def translate(self,delta,poly=False):
        if poly:
            p = Polygon(self)
            p._elem = translate(self._elem,delta)
            p._update = True
            return p

        return translate(self.geom(),delta)
    
    def geom(self):
        if self._update:
            self._updateInternals()
        return deepcopy(self._outline)
                
    def bbox(self):
        if self._update:
            self._updateInternals()
        return line(self._bbox)

    def isinside(self,p):
        if self._update:
            self._updateInternals()
        bb = self._bbox
        if not isinsidebbox(bb,p):
            return False
        p2 = add([1,1,0,1],bb[1])
        if vclose(p2,p): ## did we randomly pick an outside point near the
                         ## test point?
            p2 = sub(bb[0],[1,1,0,1])
        l = line(p,p2)
        pp = intersectGeomListXY(l,self.geom())
        if pp == False:
            return False
        return len(pp) % 2 == 1

    def grow(self,r):
        if close(r,0.0):
            return
        elif r < 0:
            raise ValueError('negative growth values not valid')
        def _growgeom(elem):
            ne = []
            for e in elem:
                if ispoint(e):
                    ne.append(arc(e,r))
                elif isline(e):
                    ne.append(arc(e[0],r))
                    ne.append(arc(e[1],r))
                elif isarc(e):
                    a = arc(e)
                    a[1][0] += r
                    ne.append(a)
                elif ispoly(e):
                    for p in e:
                        ne.append(arc(p,r))
                elif isgeomlist(e):
                    ne += _growgeom(e)
            return ne
        self._elem = _growgeom(self._elem)
        self._update=True

    def shrink(self,r):
        return False

    
## Utility functions


## function to process a geometry list and remove zero-length
## elements, including points, zero-length lines, and zero-length
## arcs.  Will recursively process geometry lists.

def cullZeroLength(geom):
    gl = []
    for g in geom:
        if ispoint(g):
            pass
        elif (isline(g) or isarc(g) or ispoly(g)) \
             and length(g) > epsilon:
            gl.append(deepcopy(g))
        elif isgeomlist(g) and length(g) > epsilon:
            gl.append(cullZeroLength(g))
        else: # zero-length or unknown element
            print("zero length element ",vstr(g)," culled")
    return gl



## make a poly containing a single circle

def makeCircle(center=point(0,0,0),radius=1.0):
    poly = Polygon()
    poly.addArc(arc(center,radius,0.0,180.0))
    poly.addArc(arc(center,radius,180.0,359.99))
    return poly

## make a rectangle with the specified width and height

def makeRect(width,height,center=point(0,0,0)):
    w=width/2.0
    h=height/2.0
    p0=add(point(-w,h),center)
    p1=add(point(-w,-h),center)
    p2=add(point(w,-h),center)
    p3=add(point(w,h),center)
    poly = Polygon()
    poly.addPoint(p0)
    poly.addPoint(p1)
    poly.addPoint(p2)
    poly.addPoint(p3)
    return poly

## make rounded rectangle with specified width, height, and chamfer,
## centered at the origin by default

def makeRoundRect(width,height,chamf,center=point(0,0,0)):
    cr=chamf/2.0
    wid=width-chamf
    hei=height-chamf
    w=wid/2.0
    h=hei/2.0
    p0=add(point(-w,h),center)
    p1=add(point(-w,-h),center)
    p2=add(point(w,-h),center)
    p3=add(point(w,h),center)
    c0=arc(p0,cr)
    c1=arc(p1,cr)
    c2=arc(p2,cr)
    c3=arc(p3,cr)
    poly = Polygon()
    poly.addArc(c0)
    poly.addArc(c1)
    poly.addArc(c2)
    poly.addArc(c3)
    return poly



## yapCAD geometry-generating superclass
## =====================================

## subclasses of Geometry implement the geom() method, which produce
## geometry lists or computational geometry primitives, as described
## by the functions in geom.py

## subclasses of SampleGeometry implelement geom() and sample()
## methods.  The geom() method is guaranteed to produce a geometry
## list, and the sample() method allows for parametric sampling, which
## is guaranteed to "make sense" in the interval 0.0 <= u <= 1.0.
## Unless there is a good reason otherwise, subclasses of
## SampleGeometry should guarantee C0 continuity*

## * For fractal elements, or other complex forms where C0 continuity
## is less meaningful, SampleGeometry might still be approprite.
## However, if you have diconnected line and arc segments, with a
## classically-definalble gap (or the possibility thereof) then
## implement Geometry instead

import copy
from yapcad.geom import *


class Geometry:
    """ generalized computational geometry class """

    def __repr__(self):
        return 'geometry base class wrapper for: {}'.format(vstr(self._elem))

    def __init__(self,a=False):
        self._update=True
        self._elem=[]
        if a:
            if ispoint(a) or isline (a) or isarc(a) or ispoly(a) \
                or isgeomlist(a) or isinstance(a,Geometry):
                self._elem=[ deepcopy(a) ]
            else:
                raise ValueError('bad argument to Geometry class constructor: {}'.format(a))
        

    def _updateInternals(self):
        return

    def geom(self):
        if self.update:
            self._updateInternals()
        return deepcopy(self._elem)

    

    
class SampleGeometry(Geometry):
    """ generalized sampleable geometry class"""

    def __init__(self,a=False):
        super().__init__(a)

    def __repr__(self):
        return 'sampleable geometry base class wrapper for: {}'.format(vstr(self._elem))
        
    def sample(self,u):
        if self._update:
            self._updateInternals()
        return sample(self.geom(),u)
    

class IntersectGeometry(SampleGeometry):
    """ generalized intersectable geometry class"""

    def __init__(self,a=False):
        super().__init__(a)

    def __repr__(self):
        return 'intersectable geometry base class wrapper for: {}'.format(vstr(self._elem))
        
    def intersectXY(self,g,inside=True,params=False):
        if self._update:
            self._updateInternals()
        return intersectXY(g,self.geom(),inside,params)
    



## yapCAD Polyline and Polygon example

print("example6.py -- yapCAD computational geometry and DXF drawing example")
print("create, sample, intersect, and draw Polygons, lines and circles")

print("""
In this demo we sample circles to find the points to create regular
polygons, then we use those points to create circular corners for
Polygon instances.  We draw those polygons in the 'PATHS' layer.

We also demonstrate calculating the intersection of a line with the
polygons in both parameter space and in XY space, and draw the
intersection points in the 'DOCUMENTATION' layer.

Finally, we sample each Polygon instance to create a series of
evenly-spaced circular 'drill holes' along the contour of the poly.
The circles increase in size to show the direction of increasing
sample parameter.  These are rendered in the 'DRILLS' layer.
""")

from yapcad.ezdxf_drawable import *
from yapcad.geom import *
from yapcad.poly import *

#set up DXF rendering
d=ezdxfDraw()
filename="example6-out"
print("Output file name is {}.dxf".format(filename))
d.filename = filename

## Put some documentary text on the drawing
d.layer = 'DOCUMENTATION'

d.draw_text("yapCAD", point(5,15),\
            attr={'style': 'OpenSans-Bold',
                  'height': 1.5})

d.draw_text("example6.py",
            point(5,12))
d.draw_text("polygons, circles, lines,",
            point(5,10))
d.draw_text("and intersections",
            point(5,8.5))
d.layer = False # back to default layer

# make circles centered at -10.0,10.0
a = point(-10.0,10.0)

# make circles centered at 10.0,-10.0
b = point(10.0,-10.0)

## We will store instances of the Polygon() class in this list
polys1 = []

## select the 'PATHS' drawing layer
d.layer = 'PATHS'

## create the first group of concentric rounded polys
for i in range(4,7): # number of sides, 4 to 6
    poly = Polygon()
    polys1.append(poly)
    ## this is the circle we will sample to get the corners of the poly
    circ1 = arc(a,i**1.4-1.0)
    for j in range(i):
        x=j/i
        p = samplearc(circ1,x) ## this is the location of the corner
        poly.addArc(arc(p,1.0)) ## add a circle of radius 1 as the corner

## create the second group of concentric rounded polys
for i in range(7,10): # number of sides, 7 to 9
    poly = Polygon()
    polys1.append(poly)
    circ2 = arc(b,(i*0.7)**1.4-1.0)
    for j in range(i):
        x=j/i
        poly.addArc(arc(samplearc(circ2,x),1.0))

## create a line between the centers of the two groups, and draw it
l = line(a,b)
d.draw(l)

## now, calculate the intersections of this line with all of the polys
## we've just created. We will do that in line parameter space as well
## as simply returning the intersection points.
pp = []
uu1s = []
for ply in polys1:
    z = intersectXY(l,ply.geom(),inside=True)
    u = intersectXY(l,ply.geom(),inside=False,params=True)
    if z:
        pp = pp + z
    if u:
        uu1s = uu1s + u[0]
    # also, draw that poly while we are at it
    d.draw(ply.geom())


## for every intersection point on the line itself, draw it as an 'x'
## in the DOCUMENTATION layer

d.polystyle='points'
d.pointstyle='x'
d.layer = 'DOCUMENTATION'
d.draw(pp)

## for every intersection in line parameter space, including
## intersections outside the 0 <= u <= 1 interval, sample that point
## and draw it as an 'o'.  This means that for the intersection points
## inside the line interval, they will be drawn as a superimposed 'o'
## on the previously drawn 'x'.

d.pointstyle='o'
for u in uu1s:
    p=sampleline(l,u)
    d.draw(p)

## Now, let's draw some evenly-spaced "drill holes" of successivly
## increasing size, tracing the outline of our polys.
d.layer = 'DRILLS' 
i = 4
for ply in polys1:
    for j in range(i*3):
        p = ply.sample(j/(i*3))
        d.draw(arc(p,0.2+j/(i*10)))
    i = i+1

d.layer = False ## Reset to the default layer -- not really necessary
d.display() ## Send the drawing to the file



## yapCAD support functions for working with Autodesk's DXF file format
## Created on Fri Oct  9 16:06:08 PDT 2020
## Copyright (c) 2020 Richard W. DeVaul
## Copyright (c) 2020 yapCAD contributors
## All rights reserved
## See licensing terms here: https://github.com/rdevaul/yapCAD/blob/master/LICENSE

from yapcad.geom import *
import math

## eqivalent to autoLISP polar function
## note this function works in radians
def polar(p,ang,r):
    x = math.cos(ang)*r
    y = math.sin(ang)*r
    return add(p,point(x,y))

## equivalent to autoLISP angle function
## note this function returns radians
def angle(p1,p2):
    if vclose(p1,p2):
        raise ValueError('coincident points')
    pp = sub(p2,p1)
    return atan2(pp[1],pp[0])

## utility functions for radian/degree conversions
def rad2deg(theta):
    return theta*360.0/pi2

def deg2rad(theta):
    return theta*pi2/360.0

## function to convert from DXF Polyline "bulge" representation to
## a yapCAD arc.
## Thanks to Lee Mac for autoLISP code examples
## http://www.lee-mac.com/bulgeconversion.html
def bulge2arc(p1,p2,b):
    a = 2.0 * math.atan(b)
    r = dist(p1,p2) / (math.sin(a)*2.0)
    c = polar(p1, ((math.pi/2.0 - a) + angle(p1, p2)),r)
    a1 = rad2deg(angle(c,p1))
    a2 = rad2deg(angle(c,p2))
    if b < 0.0:
        return arc(c,abs(r),a2,a1)
    else:
        return arc(c,abs(r),a1,a2)

## function to convert from a yapCAD arc to a DXF Polyline "bulge"
## representaiton.
## Thanks to Lee Mac for autoLISP code examples
def arc2bulge(ac):
    c = ac[0]
    r= ac[1][0]
    a1=deg2rad(ac[1][1])
    a2=deg2rad(ac[1][2])
    
    b = math.tan(((a2-a1)%pi2)/4.0)

    return [polar(c,a1,r),polar(c,a2,r),b]


## Function to convert a yapCAD poly (point list) to an internal DXF
## polyline representation
def yappoly2acpoly(plist):
    dxfpoly=[]
    for i in range(len(plist)-1):
        p=plist[i]
        dxfpoly.append([p[0],p[1],0,-2])
    p = plist[-1]
    dxfpoly.append([p[0],p[1],0,-2])
    return dxfpoly

## funtion to convert a yapCAD geometry list to a list of DXF
## polylines. Each contiguous sequence of yapCAD lines, arcs, and
## polys will be converted into a DXF polyline.

## If there are elements, such as circles, that aren't convertable to
## a DXF polyline representation, they will be added to a separate
## "nonconvertable" yapCAD geometry list.

## Return value ( dxfpolys , nonconvertable ), where dxfpolys is a
## list of contiguous dxf bulge polyline representations, and
## nonconvertable is a yapCAD geometry list of remainders

def yapgeom2acpoly(glist):
    ## internal function to take a non-zero-length geometry list and
    ## return one or more contiguous geometry lists
    def gl2contig(gl):
        nonconv=[]
        ggl=[]
        cl=[gl[0]]
        ep = sample(cl[0],1.0) # find endpoint of first element
        for i in range(1,len(gl)):
            g = gl[i]
            sp = sample(g,0.0) # find start of current element
            if vclose(sp,ep): # are they close?
                cl.append(g) # yep, append
            else:
                ggl.append(cl) #nope, finish out current list, start
                               #new one
                cl=g
            ep = sample(g,1.0)
        return ggl
    if not isgeomlist(glist):
        return False
    
    ypgls = gl2contig(glist)
    dxfgls = []
    dxfgl = []
    for ypgl in ypgls:
        dxfpoly=[]
        for i in range(1,len(ypgl)):
            if ispoint(gl):
                ## bogus, unfinished function
                return

        return


                



## geom3d, enhanced two- and three-dimensional geometry support for yapCAD
## started on Thu Oct  1 13:52:45 PDT 2020
## Richard W. DeVaul

from yapcad.geom import *

## the geometric representations of point, line, arc, poly, and
## geomlist provided by yapcad.geom are suitable for representing
## zero- or one-dimensional figures embedded in a three-dimensional
## space.  And while there is no requirement that the representations
## provided by yapcad.geom (such as an arc, line, or polyline) lie in
## the XY plane, many of the functions in yapcad.geom are explicitly
## intended to perform computational geometry operations on XY-planar
## entities. 

## Further, while a closed figure described by yapcad.geom may
## implicitly bound a two-dimensional face, there is little support
## for working with two-dimensional surfaces provided by that module.
## There is no direct support of any kind for working with
## three-dimemnsional volumes in yapcad.geom.

## in this module, we define the concept of a parametric
## two-dimensional surface and three-dimensional volume, and provide
## implicit and explicit geometry operations for working with them.

## The goal of the geom3d.yapcad module is to allow for the
## construction of two-dimensinal surfaces and three-dimensional
## geometry for the purposes of modeling, computational geometry, and
## rendering.  Specifically, we wish to support the following:

## (1) Support the implicit representation of three-dimensional
## geometry, and the performance of constructive solid geometry
## operations on this implicit geometry (union, intersection,
## difference) to produce more complex implicit three dimensional
## forms. 

## (2) Support the implicit representation of two dimensional
## surfaces, such as a planar surface specified by three points, or
## the surface of a three-dimensional object like a sphere, and allow
## for computational geometry operations on these surfaces, such as
## intersection operations, to produce explicit one-dimensional
## objects, such as lines, arcs, etc.

## (3) support the conversion of an implicit two-dimensional surface
## to an explicit, teselated triangluar geometry that may be easily
## rendered using a conventional 3D graphics rendering pipline, such
## as OpenGL

## (4) Support for the conversion of implicit three-dimenaional
## constructive solid geometry into an explicit, contiguous closed
## surface representation using the marching cubes algortihm, or any
## other user-specified conversion algoritm, for the purposes of
## interactive 3D rendering and conversion to 3D CAM formats, such as
## STL.

## Structure for 2D/3D geometry:

## geom3d in tuple( <type>, <transform>, <primary representation>,
##                  <sampled representation>, <rendering hints> )
## where:
## type in tuple ('surface' | 'solid' | 'transform', <subtype>)




## yapCAD 3D inside-outside testing example

from yapcad.geom import *
import yapcad.poly as poly
import random

def randomPoints(bbox,numpoints):
    points = []
    minx = bbox[0][0]
    maxx = bbox[1][0]
    miny = bbox[0][1]
    maxy = bbox[1][1]
    rangex = maxx-minx
    rangey = maxy-miny
    for i in range(numpoints):
        points.append(point(random.random()*rangex+minx,
                            random.random()*rangey+miny))
    return points

def pointInBox(bbox,r):
    minx = bbox[0][0]+r
    maxx = bbox[1][0]-r
    miny = bbox[0][1]+r
    maxy = bbox[1][1]-r
    rangex = maxx-minx
    rangey = maxy-miny
    x = point(random.random()*rangex+minx,
              random.random()*rangey+miny)
    return x


def randomArc(bbox,minr=0.0,maxr=10.0,circle=False):
    radr = maxr-minr
    r = random.random()*radr+minr
    start = 0
    end = 360
    if not circle:
        start = random.random()*360.0
        end = start + random.random()*360.0

    x = pointInBox(bbox,r)
    return arc(x,r,start,end)

def randomPoly(bbox,numpoints=10,minr = 1.0,maxr = 10.0):
    angles = []
    rads = []
    ang = 0.0
    rr = maxr-minr
    for i in range(numpoints):
        a = random.random()
        angles.append(ang)
        ang = ang + a
        # print("a: ",a," ang: ",ang)
        rads.append(random.random()*rr+minr)

    sf = pi2/ang

    points = []
    x = pointInBox(bbox,maxr)
    
    for i in range(numpoints):
        p = [cos(angles[i]*sf)*rads[i],
             sin(angles[i]*sf)*rads[i],0,1]
        points.append(add(p,x))

    return points + [ points[0] ]

def randomGeometry(bbox, numarcs, numcircles, numpolys):
    geom=[]
    for i in range(numarcs):
        geom.append(randomArc(bbox,minr = 5,maxr=10))

    for i in range(numcircles):
        geom.append(randomArc(bbox,minr = 3, maxr=10,circle=True))

    for i in range(numpolys):
        geom.append(randomPoly(bbox,numpoints=round(random.random()*7+3),
                               maxr =10))
    return geom

def drawGeom(dd,geom):
    for g in geom:
        if isarc(g) and not iscircle(g):
            ## draw the piza slice
            dd.draw(g)
            dd.draw(line(sample(g,1.0),
                         g[0]))
            dd.draw(line(g[0],
                         sample(g,0.0)))
        else:
            dd.draw(g)

def testPoints(points,geom,testElements=True):
    inpts=[]
    outpts=[]

    for p in points:
        ins = False
        if testElements:
            for g in geom:
                if isinsideXY(g,p):
                    inpts.append(p)
                    ins = True
                    continue
        else:
            if isinstance(geom,poly.Polygon):
                if geom.isinside(p):
                    inpts.append(p)
                    ins = True
            else:
                if isinside(geom,p):
                    inpts.append(p)
                    ins = True
        if not ins:
            outpts.append(p)
    return inpts, outpts

def testAndDraw(dd):
    dd.linecolor = 'white'

    ## this is the bounding box for our test area
    bbox = line([-60,-60,0,1],[60,60,0,1])

    ## make 2,000 points
    tps = randomPoints(bbox,2000)

    ## make 20 random arcs, 20 random circles, and 20 random polys.
    glist = randomGeometry(bbox,20,20,20)

    ## test the list of points, sort in to inside points and outside
    ## points
    inpts,outpts = testPoints(tps,glist)

    ## draw the test geometry -- we are using a special draw function
    ## that will render arcs as "pizza slices."
    drawGeom(dd,glist)

    ## Draw the points.
    dd.polystyle='points'

    dd.linecolor = 'aqua'
    dd.pointstyle = 'x'
    dd.draw(inpts)

    dd.linecolor = 'red'
    dd.pointstyle = 'o'
    dd.draw(outpts)
    dd.display()
    
if __name__ == "__main__":
    import sys
    print("example10.py -- yapCAD inside-outside testing demonstration")

    renderOgl = False
    filename="example10-out"
    oglarg= ("pyglet","opengl","OpenGL")
    dxfarg= ("ezdxf","dxf","DXF")
    if len(sys.argv) > 1:
        if sys.argv[1] in oglarg:
            renderOgl=True
        elif sys.argv[1] in dxfarg:
            renderOgl=False
        else:
            print("""
 In this example, we create different types of yapCAD geometry and
 randomly gemerate test points.  Test points that fall inside are
 rendered as aqua crosses, points that fall outside are rendered as red
 crosses.
            """)

            print("syntax: $ python3 {} <rendertype> [filename]".format(sys.argv[0]))
            print("    where <rendertype> is one of {} for OpenGL".format(oglarg))
            print("    or one of {} for DXF".format(dxfarg))
            print("    For DXF, you can optionally specify [filename].dxf as the output file")
            quit()
    if len(sys.argv) > 2 and renderOgl==False:
        filename = sys.argv[2]+".dxf"
    dd = []
    if renderOgl:
        print("OpenGL rendering selected")
        from yapcad.pyglet_drawable import *
        dd=pygletDraw()
    else:
        print("DXF rendering selected")
        from yapcad.ezdxf_drawable import *
        #set up DXF rendering
        dd=ezdxfDraw()
        dd.filename = filename
    print("rendering...")
    testAndDraw(dd)
    print("done")



## multi-rendering-back-end drawing example for yapCAD
print("example1.py -- yapCAD DXF and OpenGL drawing example")

from yapcad.ezdxf_drawable import *
from yapcad.pyglet_drawable import *
from yapcad.geom import *

#set up openGL rendering
def setupGL():
    dGl =pygletDraw()
    dGl.magnify = 1.0
    dGl.linecolor = 'white'
    dGl.cameradist = 25
    return dGl

#set up DXF rendering
def setupDXF():
    d=ezdxfDraw()
    filename="example1-out"
    print("\nOutput file name is {}.dxf".format(filename))
    d.filename = filename
    return d

## Draw some documentary text using the specified drawable.
## Text in the OpenGL rendering isn't working very well yet.

def legend(d):

    d.draw_text("yapCAD", point(5,15),\
                attr={'style': 'OpenSans-Bold', # style for ezdxf
                      'font_name': 'OpenSans', # style for pyglet
                      'bold': True, # style for pyglet
                      'height': 2.0})
    d.draw_text("example1.py",
                point(5,12))

    d.draw_text("drawing primitives",
                point(5,10))

    
## make geometry for drawing, and return it as a geometry list
def geometry():
    p = point(10,8)
    l = line(point(-5,15),
             point(10,-5))
    a = arc(arc(point(1,3),6,45,135))

    return [p,l,a]


## Do the rendering with the specified drawable
## Note: if we just wanted to render all of this the same way,
## we could just call d.draw(glist)

def drawGlist(glist,d):

    ## extract the point, line, and arc from the list
    p = glist[0]
    l = glist[1]
    a = glist[2]

    ## in this example we show three different ways of specifying the
    ## line color used for rendering.  By color index (based on the
    ## AutoCAD standard color index table), by standard HTML color
    ## name, and by specifying the RGB tripple.  All of these methods
    ## should work regardless of the drawable back-end used.
    
    ## render the point
    d.pointstyle = 'xo' # set the point rendering style
    d.linecolor = 1 # set color to red (DXF index color)
    d.draw(p)

    ## render the line in white
    d.linecolor = 'white' #set color to white, by name
    d.draw(l)

    ## render the arc in aqua, by specifying the RGB tripple.
    ## NOTE:
    ## only RGB tripples that correspond to AutoCAD color table
    ## entries will produce the desired results in DXF rendering at
    ## present.
    
    d.linecolor = [0,255,255] # corresponds to 'aqua'
    d.draw(a)

if __name__ == "__main__":

    ## setup DXF rendering
    d= setupDXF()

    ## setup OpenGL rendering
    dGl = setupGL()

    ## make the geometry
    geomlist = geometry()

    ## draw the geometry in DXF
    drawGlist(geomlist,d)

    ## add a dawing legend on the DXF drawing, in the DOCUMENTATION
    ## layer
    
    d.layer = 'DOCUMENTATION'
    d.linecolor = 256 ## set layer default color
    legend(d)

    ## add a drawing legend in the OpenGL rendering, setting the
    ## color eplicitly to yellow
    
    dGl.linecolor = 'yellow'
    legend(dGl)
    
    ## draw the geometry in OpenGL
    drawGlist(geomlist,dGl)

    ## write out the DXF file as example1-out.dxf
    d.display()

    ## create the interactive OpenGL rendering -- do this last
    dGl.display()
    



## drawing and computational geometry example for yapCAD
print("example4.py -- yapCAD computational geometry and DXF drawing example")
print("create and render rounded polygons")
print('''

This example demonstrates how to work with the yapCAD Polygon class.
The Polygon class is a generalization of the simple multi-vertex
polygon (see the poly() function in geom.py) intended to support the
creation of 'rounded' geometry and to allow the calculation of
iso-distance lines.

In this case, we create a regular pentagon, then grow that by a
distance of 2.0 to create a rounded version that represents an
iso-distance line.  Then grow it again by 4 to crate another poly that
we sample to find the centers for 11 circles that are equally spaced
along the isodistance line d = 4.0''')

from yapcad.ezdxf_drawable import *
from yapcad.geom import *
from yapcad.poly import *

#set up DXF rendering
d=ezdxfDraw()

filename="example4-out"
print("\nOutput file name is {}.dxf".format(filename))
d.filename = filename

## Put some documentary text on the drawing
d.layer = 'DOCUMENTATION'

d.draw_text("yapCAD", point(5,15),\
            attr={'style': 'OpenSans-Bold',
                  'height': 1.5})

d.draw_text("example4.py",
            point(5,12))
d.draw_text("polygons, circles, and",
            point(5,10))
d.draw_text("geometry sampling",
            point(5,8.5))
d.layer = False

# make a circle with radius 10, centered at the origin
a = point(0,0)
circ = arc(a,10.0)

# sample 5 points from the circle
points = []
for i in range(5):
    x=i/5.0
    points.append(samplearc(circ,x))

# make a circle of radius 2 at the center
center = arc(a,2.0)
d.draw(center)

# make a polygon and add five points as corners
poly0 = Polygon()
for p in points:
    poly0.addPoint(p)

# make a polygon and add five circles with radius 2.0 as corners
poly1 = Polygon()
for p in points:
    a = arc(p,2.0)
    poly1.addArc(a)

# make a third polygon and add circles with radius 4.0 as corners
poly2 = Polygon()
for p in points:
    a = arc(p,4.0)
    poly2.addArc(a)

# draw the generated geometry
d.draw(poly0.geom())
d.draw(poly1.geom())

# sample 12 points from a polygon
for i in range(12):
    x = i/12.0
    p = poly2.sample(x)
    # add some points to the drawing
    if i == 0:
        d.pointstyle = 'xo'
    else:
        d.pointstyle = 'o'
    d.layer = 'DOCUMENTATION'
    d.draw(p)
        
    d.layer = 'DRILLS'
    d.draw(arc(p,2.0))

d.display()



## yapCAD boolen operation support

from yapcad.geom import *
from yapcad.poly import *

combineDebugGL=[]

class Boolean(IntersectGeometry):
    """Boolean operations on Polygons"""

    types = ('union','intersection','difference')
    
    def __init__(self,type='union',polys=[]):
        for p in polys:
            if not ( isinstance(p,Polygon) or isinstance(p,Boolean) ):
                raise ValueError('non-poly or non-boolean passed to Boolean(): {}'.format(p))
            if not type in self.types:
                raise ValueError('invalid type passed to Boolean(): {}'.format(tpe))
            self._elem=list(polys)
            self._type=type
            self._update=True
            self._outline=[]


    def combine_geom(self,g1,g2):
        bbox1 = g1.bbox()
        bbox2 = g2.bbox()
        try :
            inter = intersectXY(g1.geom(),g2.geom(),params=True)
        except ValueError:
            print("had a problem intersecting following geometries:")
            print("g1.geom(): ",g1.geom())
            print("g2.geom(): ",g2.geom())
            raise

        if inter != False and (inter[0] == False or inter[1] == False):
            raise ValueError('bad intersection list: ',inter)
        # Given parameter values that define two potential sub-arcs of
        # a figure, determine which sub-arc is valid by looking for
        # intersections between these parameter values. In the
        # condition where u2 is smaller than u1, it's possible that
        # this represents an arc in the counter-clockwise direction
        # between u2 and u1.  It also could represent a clockwise arc
        # that "wraps around" from u1 to (u2+1).  We will check for
        # values x1, x2 in ilist that are u2 < x1 < u1 and u1 < x2 <
        # (u2+1.0).  The existance of x1 or x2 rules out the
        # corresponding arc.  If neither x1 nor x2 exists, we bias
        # towards the counter-clockwise arc.

        def between(u1,u2,ilist):
            if len(ilist) < 2:
                raise ValueError('bad ilist')
            
            if u1 > u2 :
                if len(ilist) == 2:
                    return u1,u2+1.0,False
                x1s = list(filter(lambda x: x > u2 and x < u1,ilist))
                x2s = list(filter(lambda x: x > u1 or x < u2,ilist))
                l1 = len(x1s)
                l2 = len(x2s)
                #print("u1: ",u1," u2: ",u2," ilist: ",ilist," x1s: ",x1s," x2s: ",x2s)
                if l1 > 0 and l2 > 0:
                    raise ValueError('intersections on both sides')
                elif l1 > 0:
                    print("AA")
                    if True or self._type == 'union':
                        return u1,u2+1.0,False
                    else:
                        return u2,u1,True
                else:
                    print("A-")
                    return u2,u1,True

            else:
                if len(ilist) == 2:
                    return u1,u2,False
                x1s = list(filter(lambda x: x > u1 and x < u2,ilist))
                x2s = list(filter(lambda x: x > u2 or x < u1,ilist))
                l1 = len(x1s)
                l2 = len(x2s)
                #print("u1: ",u1," u2: ",u2," ilist: ",ilist," x1s: ",x1s," x2s: ",x2s)
                if l1 > 0 and l2 > 0:
                    raise ValueError('intersections on both sides')
                elif l1 > 0:
                    print("BB")

                    if True or self._type == 'union':
                        return u2,u1+1,False
                    else:
                        return u1,u2,False
                else:
                    print("B-")
                    return u1,u2,False
            

        ## utility to perform combination on one "segment"
        def cmbin(g1,g2,itr):
            g1s = itr[0][0]
            g1e = itr[0][1]
            g2s = itr[1][0]
            g2e = itr[1][1]

            seg = []
            ZLEN1=close(g1s,g1e)
            ZLEN2=close(g2s,g2e)

            g1reverse=False
            g2reverse=False
            
            if True or self._type == 'difference':
                if g1e < g1s:
                    g1e+=1.0
                #g1s,g1e,g1reverse = between(g1s,g1e,inter[0])
                g2s,g2e,g2reverse = between(g2s,g2e,inter[1])
                g2reverse = False
            else:
                if g1e < g1s:
                    g1e+=1.0
                if g2e < g2s:
                    g2e += 1.0

            p1=g1.sample(((g1s+g1e)/2)%1.0)

            p1inside=0
            for i in range(5):
                u = (i+1)/6.0
                p = g1.sample((u*g1e+(1.0-u)*g1s)%1.0)
                if g2.isinside(p):
                    p1inside=p1inside+1

            p2inside = 0
            for i in range(5):
                u = (i+1)/6.0
                p = g2.sample((u*g2e+(1.0-u)*g2s)%1.0)
                if g1.isinside(p):
                    p2inside=p2inside+1
                
            if p1inside > 0 and p2inside > 0:
                print("warning: inside test succeeded for both p1s and p2s: ",
                      p1inside," ",p2inside)

            if p1inside == 0 and p2inside == 0:
                print("warning: inside test failed for both p1s and p2s")
                
            p2=g2.sample(((g2s+g2e)/2)%1.0)

            if ZLEN1 and ZLEN2:
                print ('both segments zero length')
                return []
            elif ZLEN2 and not ZLEN1:
                print ('zero length segment 2')
                if self._type=='union':
                    return g1.segment(g1s,g1e)
                elif self._type=='difference':
                    return 
                else: #intersection
                    return []
            elif ZLEN1 and not ZLEN2:
                print ('zero length segment 1')
                if self._type=='union':
                    if g2e < g2s:
                        g2e += 1.0
                    return g2.segment(g2s,g2e)
                else: # difference or intersection
                    return []
            
            if self._type == 'union':
                #if g2.isinside(p1):
                if p1inside > p2inside:
                    # if g2e < g2s:
                    #     g2e += 1.0
                    seg += g2.segment(g2s,g2e)
                else:
                    seg += g1.segment(g1s,g1e)                    
            elif self._type == 'intersection':
                #if g2.isinside(p1):
                if p1inside > p2inside:
                    seg += g1.segment(g1s,g1e)
                else:
                    # if g2e < g2s:
                    #     g2e += 1.0
                    #seg += g2.segment(g2s,g2e,reverse=g2reverse)
                    seg += g2.segment(g2s,g2e)
            elif self._type == 'difference':
                s = []
                #if g2.isinside(p1):
                if p1inside > p2inside:
                    pass
                else:
                    # print("rsort: ",vstr(inter))
                    seg += g1.segment(g1s,g1e)
                    # print("g2s: ",g2s," g2e: ",g2e," g2reverse: ",g2reverse)
                    s = g2.segment(g2s,g2e,reverse=g2reverse)
                    s = reverseGeomList(s)
                    seg += s
                if len(inter[0]) > 2:
                    combineDebugGL.append(s)
                    # print("seg: ",vstr(seg))

            return seg

        ## utility function to sort intersections into non-decreasing
        ## order
        def rsort(il):
            nl = []
            rl = []
            rr = []
            for i in range(len(il[0])):
                nl.append([il[0][i],il[1][i]])
            nl.sort(key=lambda x: x[0])
            for i in range(len(nl)):
                rl.append(nl[i][0])
                rr.append(nl[i][1])
            return [rl,rr]

        if inter == False: # disjoint, but bounding boxes might be
                           # null or one could be inside the other
            if not bbox1 and bbox2: # g1 is empty, but g2 contains geometry
                if self._type=='union':
                    return g2.geom()
                else:
                    return []
            if bbox1 and not bbox2: # g2 is empty, but g1 isn't
                if self._type=='union' or self._type=='difference':
                    return g1.geom()
            if not bbox1 and not bbox2: # no geometry at all
                return []
            ## OK, no intersection but it is possible that one profile
            ## could be inside the other.  Do fast bounding box checks
            ## before doing intersection-based checking.
            if isinsidebbox(bbox1,bbox2[0]) and isinsidebbox(bbox1,bbox2[1]) \
               and g1.isinside(g2.sample(0.0)): # g2 is inside g1
                ## g2 is inside g1
                if self._type == 'union':
                    return g1.geom()
                elif self._type == 'intersection':
                    return g2.geom()
                else: #difference, g2 is a hole in g1
                    return [g1.geom(),g2.geom()]
            elif isinsidebbox(bbox2,bbox1[0]) and isinsidebbox(bbox2,bbox1[1]) \
                 and g2.isinside(g1.sample(0.0)): # g1 is inside g2
                ## g1 is indside g2
                if self._type == 'union':
                    return g2.geom()
                elif self._type == 'intersection':
                    return g1.geom()
                else: #difference, g2 has eaten g1
                    return []
            else: # g1 and g2 are disjoint
                if self._type == 'union':
                    return [g1.geom(),g2.geom()]
                elif self._type == 'difference':
                    return g1.geom()
                else: #intersection
                    return []
        if len(inter[0]) == 1 and len(inter[1]) == 1:
        ## single point of intersection:
            if self._type == 'union':
                return [g1.geom(), g2.geom()]
            elif self._type == 'difference':
                return g1.geom()
            else: #intersection
                return []
        ## There are two or more points of intersection.
        inter = rsort(inter)
        #print("rsort: ",vstr(inter))

        if len(inter[0]) %2 != 0:
            print("WARNING: odd number of intersections (",len(inter[0]),", unpredictable behavior may result")
        r = []
        for i in range(1,len(inter[0])+1):
            r += cmbin(g1,g2,[[inter[0][i-1],
                               inter[0][i%len(inter[0])]],
                              [inter[1][i-1],
                               inter[1][i%len(inter[1])]]])
        return r

    def bbox(self):
        return bbox(self.geom())

    def getCenter(self):
        gl = self.geom()
        if gl == []:
            raise ValueError('empty Boolean, no center')
        return center(gl)

    def getLength(self):
        gl = self.geom()
        if gl == []:
            raise ValueError('empty Boolean, no length')
        return length(gl)

    def segment(self,u1,u2,reverse=False):
        gl = self.geom()
        if gl == []:
            raise ValueError('empty Boolean, segment not defined')
        return segmentgeomlist(gl,u1,u2,closed=True,reverse=reverse)

    def mirror(self,plane,poly=False):
        b = deepcopy(self)
        b._elem = []
        b._update=True
        for p in self._elem:
            p2 = p.mirror(plane,poly=True)
            b._elem.append(p2)
        if poly:
            return b
        return b.geom()

    def rotate(self,angle,cent=point(0,0,0),axis=point(0,0,1),poly=False):
        b = deepcopy(self)
        b._elem = []
        b._update=True
        for p in self._elem:
            p2 = p.rotate(angle,cent,axis,poly=True)
            b._elem.append(p2)
        if poly:
            return b
        return b.geom()

    def scale(self,sx,sy=False,sz=False,cent=point(0,0),poly=False):
        b = deepcopy(self)
        b._elem = []
        b._update=True
        for p in self._elem:
            p2 = p.scale(sx,sy,sz,cent,poly=True)
            b._elem.append(p2)
        if poly:
            return b
        return b.geom()                        

    def translate(self,delta,poly=False):
        b = deepcopy(self)
        b._elem = []
        b._update=True
        for p in self._elem:
            p2 = p.translate(delta,poly=True)
            b._elem.append(p2)
        if poly:
            return b
        return b.geom()                        
            
    def sample(self,u):
        gl = self.geom()
        if gl == []:
            raise ValueError('empty Boolean, sample not defined')
        return sample(gl,u)

    def isinside(self,p):
        gm = self.geom()
        if gm == []:
            raise ValueError('empty Boolean, inside not defined')
        bb = bbox(gm)
        p2 = add([1,1,0,1],bb[1])
        l = line(p,p2)

        pp = intersectGeomListXY(l,gm)
        if pp == False:
            return False
        return len(pp) % 2 == 1

    def grow(self,r):
        if close(r,0.0):
            return
        elif r < 0:
            raise ValueError('negative growth not allowed')

        if self._type in ['union','intersection']:
            for p in self._elem:
                p.grow(r)
            self._update = True
        else:
            raise NotImplementedError("Don't have grow support for {} yet".format(self._type))
        

    def geom(self):
        if self._update:
            if len(self._elem)==2:
                self._outline = self.combine_geom(self._elem[0],self._elem[1])
                self._outline = cullZeroLength(self._outline)
                self._update = False
            else:
                raise NotImplementedError("don't know how to do {} yet for {} polygons".format(self._type,len(self._elem)))
        return deepcopy(self._outline)
        
        



## base class of drawable for yapCAD
## Copyright (c) 2020 Richard W. DeVaul
## Copyright (c) 2020 yapCAD contributors
## All rights reserved
## See licensing terms here: https://github.com/rdevaul/yapCAD/blob/master/LICENSE

from yapcad.geom import *

## Generic drawing functions -- assumed to use current coordinate
## transform and drawing pen (color, line weight, etc.)
 
class Drawable:
    """Base class for yapCAD drawables"""

    ## utility functions for drawing
    ## -----------------------------
    
    ## pure virtual functions -- override for specific rendering
    ## system
    def draw_arc(self,p,r,start,end):
        print("pure virtual draw_arc called: {}, {}, {}, {}".format(p,r,start,end)) 
        return

    def draw_line(self,p1,p2):
        print("pure virtual draw_line called: {}, {}".format(p1,p2))
        return

    def draw_text(self,text,location,
                  align='left',
                  attr={}):
        print("pure virtual draw_text called: {}, {}, {}, {}".format(text,location,align,attr))
        return

    ## non-virtual utility drawing functions 
    def draw_circle(self,p,r):
        self.draw_arc(p,r,0.0,360.0)
        return

    ## utility function to draw an "x", centered on point p inside square of
    ## dimension d
    def draw_x(self,p,d):
        hd=d/2
        p1=add(p,vect([-hd,-hd,0]))
        p2=add(p,vect([hd,hd,0]))
        p3=add(p,vect([-hd,hd,0]))
        p4=add(p,vect([hd,-hd,0]))
        self.draw_line(p1,p2)
        self.draw_line(p3,p4)
        return;

    ## utiitly function to draw a point based on the current point style
    def draw_point(self,p):
        if 'x' in self.__pointstyle:
            self.draw_x(p,self.__pointsize*2)

        if 'o' in self.__pointstyle:
            self.draw_circle(p,self.__pointsize)

    def __init__(self):
        self.__pointstyle = 'xo'
        self.__pointsize = 0.1
        self.__linetype = False
        self.__linewidth = 0.1
        self.__linecolor = False
        self.__fillcolor = False
        self.__polystyle = 'lines'
        self.__layer = False


    ## Various property functions

    @property
    def layer(self):
        return self.__layer
    
    def _set_layer(self,lyr):
        self.__layer = lyr
        
    @layer.setter
    def layer(self,lyr=False):
        if lyr == False or lyr == 'default':
            self._set_layer(lyr)
        else:
            raise ValueError('bad layer: ' + str(lyr))
        
    @property
    def polystyle(self):
        return self.__polystyle

    def _set_polystyle(self,pst):
        self.__polystyle = pst
        
    @polystyle.setter
    def polystyle(self,pst=False):
        if pst in [False, 'points','lines','both']:
            self._set_polystyle(pst)
        else:
            raise ValueError('bad polystyle')
        
    @property
    def pointstyle(self):
        return self.__pointstyle

    def _set_pointstyle(self,pst):
        self.__pointstyle=pst
    
    @pointstyle.setter
    def pointstyle(self,pst=False):
        if pst in [False, 'x','o','xo']:
            self._set_pointstyle(pst)
        else:
            raise ValueError('bad pointstyle: ' + str(pst))

    @property
    def pointsize(self):
        return self.__pointsize

    def _set_pointsize(self,ps):
        self.__pointsize=ps
        
    @pointsize.setter
    def pointsize(self,ps=False):
        if not isinstance(ps,(int,float)):
            raise ValueError('invalid pointsize ' + str(ps))
        if isinstance(ps,bool) and ps == False:
            self._set_pointsize(0.1)
        elif ps < epsilon:
            ps = epsilon
        self._set_pointsize(ps)

    @property
    def linewidth(self):
        return self.__linewidth

    def _set_linewidth(self,lw):
        self.__linewidth=lw

    @linewidth.setter
    def linewidth(self,lw=False):
        if not isinstance(lw,(int,float)):
            raise ValueError('invalid pointsize ' + str(lw))
        if isinstance(lw,bool) and lw ==False:
            lw = 0.1
        elif lw < epsilon:
            lw = epsilon
        self._set_linewidth(lw)

    @property
    def linetype(self):
        return self.__linetype

    def _set_linetype(self,lt):
        self.__linetype = lt

    @linetype.setter
    def linetype(self,lt=False):
        if lt in [ False, 'Continuous' ]:       # only continuous linetype supported in base class
            self._set_linetype(False)
        else:
            raise ValueError('unsupported linetype ' + str(lt))

    ## color is a complex property -- it can be set as an AUTOCAD
    ## index color, a standard color name, or an RGB tripple

    def __checkcolor(self,c):
        def isbyte(x):
            return isinstance(x,int) and x >= 0 and x <= 255
        if isinstance(c,bool) and c == False:
            return True
        elif (isinstance(c,list) and len(c) == 3 \
              and isbyte(c[0]) and isbyte(c[1]) and isbyte(c[2])) or \
             (isinstance(c,int) and c >= 0 and c < len(self.colormapAUTOCAD)) or\
             isinstance(c,str) and c in self.colordict:
            return True
        return False
    
    ## line color
    @property
    def linecolor(self):
        return self.__linecolor

    def _set_linecolor(self,c):
        self.__linecolor=c

    @linecolor.setter
    def linecolor(self,c=False):
        if self.__checkcolor(c):
            self._set_linecolor(c)
        else:
            raise ValueError('bad linecolor ' + str(c))

    ## fill color
    @property
    def fillcolor(self):
        return self.__fillcolor

    def _set_fillcolor(self,c):
        self.__fillcolor = c

    @fillcolor.setter
    def fillcolor(self,c=False):
        if self.__checkcolor(c):
            self._set_fillcolor(c)
        else:
            raise ValueError('bad fillcolor ' + str(c))

    ## non-property methods
    
    def __repr__(self):
        return 'an abstract Drawable instance'

    def str(self):
        return self.__repr__()
    
    def print(self):
        print(self.str())

    def draw(self,x):
        if ispoint(x):
            self.draw_point(x)
        elif isline(x):
            self.draw_line(x[0],x[1])
        elif isarc(x):
            self.draw_arc(x[0],x[1][0],x[1][1],x[1][2])
        elif ispoly(x):
            if self.polystyle in ('points','lines','both'):
                if self.polystyle == 'points' or \
                   self.polystyle == 'both':
                    for e in x:
                        self.draw(e)
                if self.polystyle == 'lines' or \
                   self.polystyle == 'both':
                    for i in range(1,len(x)):
                        self.draw(line(x[i-1],x[i]))
            else:
                raise ValueError("bad value for polystyle: {}".format(self.polystyle))
            
        elif isgeomlist(x):
            for e in x:
                self.draw(e)
        else:
            raise ValueError('bad argument to Drawable.draw(): '.format(x))
        
    ## cause drawing page to be rendered -- pure virtual in base class
    def display(self):
        print('pure virtual display function called')
        return True


    ## Standard color names, ala HTML: https://www.rapidtables.com/web/color/RGB_Color.html
    colordict= {
        'black':   [ [0,0,0], 0],
        'white':   [ [255,255,255], 7],
        'red':     [ [255,0,0], 1],
        'lime':    [ [0,255,0],3],
        'blue':    [ [0,0,255], 5],
        'yellow':  [ [255,255,0], 2],
        'silver':  [ [192,192,192], 9],
        'gray':    [ [128,128,128], 8],
        'maroon':  [ [127,0,0], 14],
        'olive':   [ [127,127,0], 54],
        'green':   [ [0,127,0], 94],
        'aqua':    [ [0,255,255], 4],
        'teal':    [ [0,127,127], 134],
        'navy':    [ [0,0,128], 174],
        'fuchsia': [ [255,0,255], 6],
        'purple':  [ [128,0,128], 214] }
    
    ## AUTOCAD color index:
    ## https://forums.autodesk.com/t5/visual-lisp-autolisp-and-general/
    ## rgb-values-of-the-256-standard-indexed-colors/td-p/848912
    colormapAUTOCAD = [
        [0,0,0],
	[255,0,0],
	[255,255,0],
	[0,255,0],
	[0,255,255],
	[0,0,255],
	[255,0,255],
	[255,255,255],
	[128,128,128],
	[192,192,192],

	[255,0,0],
	[255,127,127],
	[165,0,0],
	[165,82,82],
	[127,0,0],
	[127,63,63],
	[76,0,0],
	[76,38,38],
	[38,0,0],
	[38,19,19],

	[255,63,0],
	[255,159,127],
	[165,41,0],
	[165,103,82],
	[127,31,0],
	[127,79,63],
	[76,19,0],
	[76,47,38],
	[38,9,0],
	[38,23,19],

	[255,127,0],
	[255,191,127],
	[168,82,0],
	[165,124,82],
	[127,63,0],
	[127,95,63],
	[76,38,0],
	[76,57,38],
	[38,19,0],
	[38,28,19],

	[255,191,0],
	[255,223,127],
	[165,124,0],
	[165,145,82],
	[127,95,0],
	[127,111,63],
	[76,57,0],
	[76,66,38],
	[38,28,0],
	[38,33,19],

	[255,255,0],
	[255,255,127],
	[165,165,0],
	[165,165,82],
	[127,127,0],
	[127,127,63],
	[76,76,0],
	[76,76,38],
	[38,38,0],
	[38,38,19],

	[191,255,0],
	[223,255,127],
	[124,165,0],
	[145,165,82],
	[95,127,0],
	[111,127,63],
	[57,76,0],
	[66,76,38],
	[28,38,0],
	[33,38,19],

	[127,255,0],
	[191,255,127],
	[82,165,0],
	[124,165,82],
	[63,127,0],
	[95,127,63],
	[38,76,0],
	[57,76,38],
	[19,38,0],
	[28,38,19],

	[63,255,0],
	[159,255,127],
	[41,165,0],
	[103,165,82],
	[31,127,0],
	[79,127,63],
	[19,76,0],
	[47,76,38],
	[9,38,0],
	[23,38,19],

	[0,255,0],
	[127,255,127],
	[0,165,0],
	[82,165,82],
	[0,127,0],
	[63,127,63],
	[0,76,0],
	[38,76,38],
	[0,38,0],
	[19,38,19],

	[0,255,63],
	[127,255,159],
	[0,165,41],
	[82,165,103],
	[0,127,31],
	[63,127,79],
	[0,76,19],
	[38,76,47],
	[0,38,9],
	[19,38,23],

	[0,255,127],
	[127,255,191],
	[0,165,82],
	[82,165,124],
	[0,127,63],
	[63,127,95],
	[0,76,38],
	[38,76,57],
	[0,38,19],
	[19,38,28],

	[0,255,191],
	[127,255,223],
	[0,165,124],
	[82,165,145],
	[0,127,95],
	[63,127,111],
	[0,76,57],
	[38,76,66],
	[0,38,28],
	[19,38,33],

	[0,255,255],
	[127,255,255],
	[0,165,165],
	[82,165,165],
	[0,127,127],
	[63,127,127],
	[0,76,76],
	[38,76,76],
	[0,38,38],
	[19,38,38],

	[0,191,255],
	[127,223,255],
	[0,124,165],
	[82,145,165],
	[0,95,127],
	[63,111,127],
	[0,57,76],
 	[38,66,76],
	[0,28,38],
	[19,33,38],

	[0,127,255],
	[127,191,255],
	[0,82,165],
	[82,124,165],
	[0,63,127],
	[63,95,127],
	[0,38,76],
	[38,57,76],
	[0,19,38],
	[19,28,38],

	[0,63,255],
	[127,159,255],
	[0,41,165],
	[82,103,165],
	[0,31,127],
	[63,79,127],
	[0,19,76],
	[38,47,76],
	[0,9,38],
	[19,23,38],

	[0,0,255],
	[127,127,255],
	[0,0,165],
	[82,82,165],
	[0,0,127],
	[63,63,127],
	[0,0,76],
	[38,38,76],
	[0,0,38],
	[19,19,38],

	[63,0,255],
	[159,127,255],
	[41,0,165],
	[103,82,165],
	[31,0,127],
	[79,63,127],
	[19,0,76],
	[47,38,76],
	[9,0,38],
	[23,19,38],

	[127,0,255],
	[191,127,255],
	[82,0,165],
	[124,82,165],
	[63,0,127],
	[95,63,127],
	[38,0,76],
	[57,38,76],
	[19,0,38],
	[28,19,38],

	[191,0,255],
	[223,127,255],
	[124,0,165],
	[145,82,165],
	[95,0,127],
	[111,63,127],
	[57,0,76],
	[66,38,76],
	[28,0,38],
	[33,19,38],

	[255,0,255],
	[255,127,255],
	[165,0,165],
	[165,82,165],
	[127,0,127],
	[127,63,127],
	[76,0,76],
	[76,38,76],
	[38,0,38],
	[38,19,38],

	[255,0,191],
	[255,127,223],
	[165,0,124],
	[165,82,145],
	[127,0,95],
	[127,63,111],
	[76,0,57],
	[76,38,66],
	[38,0,28],
	[38,19,33],

	[255,0,127],
	[255,127,191],
        [165,0,82],
	[165,82,124],
	[127,0,63],
	[127,63,95],
	[76,0,38],
	[76,38,57],
	[38,0,19],
	[38,19,28],

	[255,0,63],
	[255,127,159],
	[165,0,41],
	[165,82,103],
	[127,0,31],
	[127,63,79],
	[76,0,19],
	[76,38,47],
	[38,0,9],
	[38,19,23],

	[0,0,0],
	[45,45,45],
	[91,91,91],
	[137,137,137],
	[183,183,183],
	[179,179,179],
        False                   # bylaer
    ]

    ## function to convert between different color representations 
    def thing2color(self,thing,convert='b',colormap=None,colordict=None):
        def _b2f(c):
            return [ c[0]/255.0,c[1]/255.0,c[2]/255.0 ]
        def _f2b(c):
            return [ round(c[0]*255.0),rount(c[1]*255.0),round(c[2]*255.0) ]
        def _b2i(c):
            for i in range(len(colormap)):
                if colormap[i] == thing:
                    return i
            return False
        def _f2i(c):
            return _b2i(_f2b(c))
        def _isgoodf(x):
            return isinstance(x,float) and x >= 0.0 and x <= 1.0
        def _isgoodb(x):
            return isinstance(x,int) and x >= 0 and x < 256
        def _isgoodfc(c):
            return isinstance(c,(list,tuple)) and len(c) == 3 and\
                _isgoodf(c[0]) and _isgoodf(c[1]) and _isgoodf(c[2])
        def _isgoodbc(c):
            return isinstance(c,(list,tuple)) and len(c) == 3 and\
                _isgoodb(c[0]) and _isgoodb(c[1]) and _isgoodb(c[2])

        if colormap == None:
            colormap = self.colormapAUTOCAD
        if colordict == None:
            colordict = self.colordict
        if convert not in ['f','b','i']:
            raise ValueError('bad colormap conversion')
        c = False
        if _isgoodfc(thing):
            if convert == 'b':
                return _f2b(thing)
            elif convert == 'i':
                return _f2i(thing)
            else:
                return thing
        elif _isgoodbc(thing):
            if convert == 'i':
                return _b2i(thing)
            if convert == 'f':
                return _b2f(thing)
            else:
                return thing
        elif isinstance(thing,str):
            cd = colordict[thing]
            if isinstance(cd,bool) and cd == False:
                raise ValueError('bad color name passed to thing2color: {}'.format(thing))
            if convert == 'i':
                return cd[1]
            else:
                c = cd[0]
        elif isinstance(thing,int):
            if thing < 0 or thing > len(colormap):
                raise ValueError('colormap index out of range: {}'.format(thing))
            if convert == 'i':
                return thing
            else:
                c = colormap[thing]
        else:
            raise ValueError('bad thing passed to thing2color: {}'.format(thing))
        if isinstance(c,bool) and c == False:
            return False
        if convert == 'i':
            return _b2i(c)
        elif convert == 'f':
            return _b2f(c)
        return list(c)
    

